from json import dumps, loads
from typing import List

import numpy as np
from tqdm import tqdm

from examtool.api.scramble import scramble
from examtool.api.watermarks import Point, get_watermark_points


def decode_watermark(exam_data, roster, corners: List[Point], bits: List[Point]):
    """
    Assume x coord increases from left to right, y increases from top to bottom
    """
    observed_points = correct_watermark_bits(corners, bits)
    email_distances = [
        [email, bit_distance(observed_points, exam_data, email)]
        for email, _ in tqdm(roster)
    ]
    email_distances.sort(key=lambda x: x[1])
    return email_distances[:10]


def bit_distance(observed_points: List[Point], exam_data, email):
    scrambled_exam = scramble(email, loads(dumps(exam_data)))
    expected_points = get_watermark_points(scrambled_exam["watermark"]["value"])
    assert len(observed_points) >= len(expected_points) / 2, "Too few observed bits"
    costs = []
    for observed_point in observed_points:
        closest = min(expected_points, key=observed_point.dist)
        costs.append(
            observed_point.dist(closest) ** 2
        )  # large penalty for misalignment
    costs.sort()
    del costs[-5:]
    return sum(costs) / len(costs)


def correct_watermark_bits(corners: List[Point], bits: List[Point]):
    assert len(corners) == 4, "All four corners must be selected"
    corners.sort(key=lambda pt: pt.x)
    left, right = corners[:2], corners[2:]
    left.sort(key=lambda pt: pt.y)
    right.sort(key=lambda pt: pt.y)
    top_left, bottom_left = left
    top_right, bottom_right = right

    def compute_basis_map(points):
        *base, ref = points
        A = np.array(
            [
                *zip(*base),
                [1, 1, 1],
            ]
        )
        b = np.array([*ref, 1])
        coeffs = np.linalg.solve(A, b)
        return A * coeffs

    source_transform = compute_basis_map(
        [
            top_left,
            top_right,
            bottom_right,
            bottom_left,
        ]
    )

    target_transform = compute_basis_map(
        [Point(5, 5), Point(105, 5), Point(105, 105), Point(5, 105)]
    )

    full_transform = target_transform @ np.linalg.inv(source_transform)

    homogenized_bits = full_transform @ np.vstack(
        [np.array(list(zip(*bits))), np.ones(len(bits))]
    )

    return [
        Point(x % 100, y % 100)
        for x, y in zip(
            homogenized_bits[0] / homogenized_bits[2],
            homogenized_bits[1] / homogenized_bits[2],
        )
    ]



import unittest

import gui_files.svg as svg


class TestSVG(unittest.TestCase):
    def assert_contains_str(self, str, substr):
        self.assertTrue(str.find(substr) > -1, "%s does not contain %s" % (str, substr))

    def test_create_graphic(self):
        graphic = svg.create_graphic(200, 300)
        self.assertEqual(
            str(graphic),
            """<svg width="200" height="300" xmlns="http://www.w3.org/2000/svg"></svg>""",
        )

    def test_draw_rect(self):
        graphic = svg.create_graphic(200, 300)
        svg.draw_rect(graphic, 10, 20, 30, 40)
        self.assert_contains_str(
            str(graphic),
            """<rect x="10" y="20" width="30" height="40" stroke="black" fill="black" />""",
        )

    def test_draw_circle(self):
        graphic = svg.create_graphic(200, 300)
        svg.draw_circle(graphic, 10, 20, 30)
        self.assert_contains_str(
            str(graphic),
            """<circle cx="10" cy="20" r="30" stroke="black" fill="black" />""",
        )

    def test_draw_line(self):
        graphic = svg.create_graphic(200, 300)
        svg.draw_line(graphic, 10, 20, 30, 40)
        self.assert_contains_str(
            str(graphic), """<line x1="10" y1="20" x2="30" y2="40" stroke="black" />"""
        )

    def test_draw_triangle(self):
        graphic = svg.create_graphic(200, 300)
        svg.draw_triangle(graphic, 10, 20, 15, 150, 150, 150)
        self.assert_contains_str(
            str(graphic),
            """<polygon points="10,20 15,150 150,150" stroke="black" fill="black"/>""",
        )

    def test_write_text(self):
        graphic = svg.create_graphic(200, 300)
        svg.write_text(graphic, 10, 20, "Turn over")
        self.assert_contains_str(
            str(graphic),
            """<text x="10" y="20" stroke="black" fill="black" font-size="medium" font-family="serif">Turn over</text>""",
        )

    def test_write_text_font(self):
        graphic = svg.create_graphic(200, 300)
        svg.write_text(
            graphic, 10, 20, "Turn over", font_size="20", font_family="sans-serif"
        )
        self.assert_contains_str(
            str(graphic),
            """<text x="10" y="20" stroke="black" fill="black" font-size="20" font-family="sans-serif">Turn over</text>""",
        )

    def test_stroke_fill(self):
        graphic = svg.create_graphic(200, 300)
        svg.draw_rect(graphic, 10, 20, 30, 40, "pink", "blue")
        self.assert_contains_str(
            str(graphic),
            """<rect x="10" y="20" width="30" height="40" stroke="pink" fill="blue" />""",
        )


if __name__ == "__main__":
    unittest.main()



import json

from tqdm import tqdm

from examtool.api.database import get_exam, get_roster, get_submissions
from examtool.api.extract_questions import extract_questions
from examtool.api.scramble import scramble


def download(exam, emails_to_download: [str] = None, debug: bool = False):
    exam_json = get_exam(exam=exam)
    exam_json.pop("secret")
    exam_json = json.dumps(exam_json)

    template_questions = list(extract_questions(json.loads(exam_json)))

    total = [
        ["Email"]
        + [question["text"] for question in extract_questions(json.loads(exam_json))]
    ]

    email_to_data_map = {}

    if emails_to_download is None:
        roster = get_roster(exam=exam)
        emails_to_download = [email for email, _ in roster]

    i = 1
    for email, response in tqdm(
        get_submissions(exam=exam), dynamic_ncols=True, desc="Downloading", unit="Exam"
    ):
        i += 1
        if emails_to_download is not None and email not in emails_to_download:
            continue

        if debug and 1 < len(response) < 10:
            tqdm.write(email, response)

        total.append([email])
        for question in template_questions:
            total[-1].append(response.get(question["id"], ""))

        student_questions = list(
            extract_questions(scramble(email, json.loads(exam_json), keep_data=True))
        )

        email_to_data_map[email] = {
            "student_questions": student_questions,
            "responses": response,
        }

    return json.loads(exam_json), template_questions, email_to_data_map, total


def get_question_to_page_mapping(
    template_questions,
    exam,
    out,
    name_question,
    sid_question,
    dispatch=None,
):
    questions = []
    pages = []
    for q in tqdm(
        template_questions,
        desc="Getting question page numbers",
        unit="Question",
        dynamic_ncols=True,
    ):
        questions.append(q)
        pdf = write_exam(
            None,
            {},
            exam,
            questions,
            questions,
            name_question,
            sid_question,
            dispatch,
        )
        pages.append(pdf.page_no())
    # for i, q in enumerate(questions):
    #     print(f"[{i}] pg: {pages[i]} - {q['id']}")
    # import ipdb; ipdb.set_trace()
    return pages



import requests
import json

GRADESCOPE_URL = (
    "https://www.gradescope.com/api/v1/courses/{}/assignments/{}/submissions"
)


class APIClient:
    def __init__(self):
        self.session = requests.Session()
        self.token = None

    def log_in(self, email, password):
        login_url = "https://www.gradescope.com/api/v1/user_session"

        form_data = {"email": email, "password": password}
        r = self.post(login_url, data=form_data)
        # print(r.status_code, r.json())

        self.token = r.json()["token"]

    def is_logged_in(self):
        return self.token is not None

    def post(self, *args, **kwargs):
        return self.session.post(*args, **kwargs)

    def upload_submission(self, course_id, assignment_id, student_email, filename):
        url = GRADESCOPE_URL.format(course_id, assignment_id)
        form_data = {"owner_email": student_email}
        files = {"pdf_attachment": open(filename, "rb")}
        request_headers = {"access-token": self.token}

        num_attempts = 0
        while num_attempts < 5:  # Control how many times to retry
            r = self.post(url, data=form_data, headers=request_headers, files=files)
            if r.ok:
                return True
            num_attempts += 1

        # Report error
        print(
            f"Issue uploading to Gradescope ({r.status_code}). Gradescope error output below:",
            student_email,
        )
        content = json.loads(r._content)
        error_lines = content.get("errors")
        if error_lines is None:
            print(content)
        else:
            for line in error_lines:
                print(line)
        print("Upload URL:", url)
        return False



import base64
import os
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

password = input("Password? ").encode("ascii")
salt = os.urandom(16)
kdf = PBKDF2HMAC(
    algorithm=hashes.SHA256(),
    length=32,
    salt=salt,
    iterations=100000,
    backend=default_backend(),
)
key = base64.urlsafe_b64encode(kdf.derive(password))
print("Key: ", key)



import json
import os
import pathlib

import click

from examtool.api.database import get_full_logs, get_roster, get_logs
from examtool.cli.utils import exam_name_option, hidden_output_folder_option


@click.command()
@exam_name_option
@hidden_output_folder_option
@click.option(
    "--full", default=False, is_flag=True, help="Get keylogging logs too, if available."
)
@click.option(
    "--fetch_all",
    default=False,
    is_flag=True,
    help="Re-download all logs.",
)
def save_logs(exam, out, full, fetch_all):
    """
    Save the full submission log for later analysis.
    Note that this command is slow.
    To view a single log entry, run `examtool log`.
    """
    out = out or "out/logs/" + exam
    full_out = os.path.join(out, "full")

    pathlib.Path(out).mkdir(parents=True, exist_ok=True)
    pathlib.Path(full_out).mkdir(parents=True, exist_ok=True)

    roster = get_roster(exam=exam)
    for i, (email, deadline) in enumerate(roster):
        print(email)
        try:
            target = os.path.join(out, email)
            if os.path.exists(target) and not fetch_all:
                print("Skipping", email)
            else:
                print("Fetching short logs for", email)
                logs = get_logs(exam=exam, email=email)
                with open(target, "w") as f:
                    json.dump(logs, f)
            if full:
                target = os.path.join(full_out, email)
                if os.path.exists(target) and not fetch_all:
                    print("Skipping", email, "for full logs")
                else:
                    print("Fetching full logs for", email)
                    logs = get_full_logs(exam=exam, email=email)
                    with open(os.path.join(full_out, email), "w") as f:
                        json.dump(logs, f)
        except KeyboardInterrupt:
            raise
        except:
            print("Failure for email", email, "continuing...")



import json
from dataclasses import dataclass
from typing import Dict

from examtool.api.utils import dict_to_list
from tqdm import tqdm

from examtool.api.database import get_exam, get_roster
from examtool.api.extract_questions import extract_questions, get_name
from examtool.api.scramble import scramble, get_elements


@dataclass
class SuspectedCheating:
    question: str
    email: str
    base_keyword: str
    expected: str
    observed: str
    answer: str
    substitutions: Dict[str, str]

    def explain(self):
        print(
            f"In question {self.question}, student {self.email} used keyword {self.observed} for "
            f"{self.base_keyword}, when they should have used {self.expected}"
        )

        print(
            f"\tThey wrote `{' '.join(self.answer.split())}`. Their substitutions were: {self.substitutions}"
        )


def find_unexpected_words(exam, logs):
    data = get_exam(exam=exam)
    exam_json = json.dumps(data)
    original_questions = {q["id"]: q for q in extract_questions(json.loads(exam_json))}
    suspected_cheating = []
    for i, (email, log) in enumerate(tqdm(logs)):
        all_alternatives = get_substitutions(data)
        scrambled_questions = {
            q["id"]: q
            for q in extract_questions(
                scramble(email, json.loads(exam_json), keep_data=True), nest_all=True
            )
        }
        flagged_question_variants = set()
        for record in log:
            record.pop("timestamp")
            for question, answer in record.items():
                question = question.split("|")[0]
                if question not in all_alternatives:
                    continue

                student_substitutions = scrambled_questions[question]["substitutions"]

                for keyword in student_substitutions:
                    for variant in all_alternatives[question][keyword]:
                        if variant == student_substitutions[keyword]:
                            continue
                        if (question, keyword, variant) in flagged_question_variants:
                            continue
                        if variant in answer:
                            # check for false positives
                            if variant in scrambled_questions[question]["text"]:
                                continue

                            flagged_question_variants.add((question, keyword, variant))

                            suspected_cheating.append(
                                SuspectedCheating(
                                    get_name(original_questions[question]),
                                    email,
                                    keyword,
                                    student_substitutions[keyword],
                                    variant,
                                    answer,
                                    student_substitutions,
                                )
                            )

    return suspected_cheating


def find_keyword(exam, phrase):
    data = get_exam(exam=exam)
    exam_json = json.dumps(data)
    for email, _ in get_roster(exam=exam):
        scrambled = scramble(email, json.loads(exam_json))
        if phrase in json.dumps(scrambled):
            print(email)


def get_substitutions(exam):
    out = {}

    def process_element(element):
        substitutions = element.get("substitutions", {}).copy()
        for item in element.get("substitutions_match", []):
            for directive in item["directives"]:
                substitutions[directive] = item["replacements"]
        for blocks in element.get("substitution_groups", []):
            directives = blocks["directives"]
            replacements = dict_to_list(blocks["replacements"])
            for directive, directive_replacements in zip(
                directives, zip(*(dict_to_list(d) for d in replacements))
            ):
                substitutions[directive] = directive_replacements
        for key in element.get("substitution_ranges", {}):
            substitutions[key] = []  # hack to make RANGEs not crash
        return substitutions

    def process_group(group, substitutions):
        group_substitutions = process_element(group)
        for element in get_elements(group):
            if element.get("type") == "group":
                process_group(element, {**substitutions, **group_substitutions})
            else:
                process_question(element, {**substitutions, **group_substitutions})

    def process_question(question, substitutions):
        out[question["id"]] = {**substitutions, **process_element(question)}

    global_substitutions = process_element(exam)
    for group in exam["groups"]:
        process_group(group, global_substitutions)

    return out



"""
Developed by ThaumicMekanism [Stephan K.] - all credit goes to him!
"""
import os
import json

import click

from examtool.cli.utils import hidden_target_folder_option
from examtool.api.gradescope_autograde import GradescopeGrader


@click.command()
@click.option(
    "--exam",
    default="cs61a-test-final",
    prompt=True,
    help="The list of exam names. If it is just one exam, just include one. Separate exams with commas: Eg 'cs61a-final-8am,cs61c-final-10am`",
)
@click.option(
    "--name-question",
    default=None,
    prompt=True,
    help="The ID of the question for the student's name.",
)
@click.option(
    "--sid-question",
    default=None,
    prompt=True,
    help="The ID of the question for the student's SID.",
)
@click.option("--course", prompt=True, help="The Gradescope course ID.")
@click.option(
    "--assignment",
    default=None,
    help="The Gradescope assignment ID. If this is left blank, this tool will create the Gradescope assignment.",
)
@click.option(
    "--assignment-title",
    default="Examtool Exam",
    help="The title you want the Gradescope assignment to have.",
)
@click.option("--email", help="Your Gradescope email address.")
@click.option("--password", hide_input=True, help="Your Gradescope account password.")
@click.option(
    "--token",
    "-t",
    default=None,
    help="The path to the token file holding your Gradescope credentials.",
    type=click.Path(),
)
@click.option(
    "--emails",
    default=None,
    help="This is a list of emails you want to autograde to the assignment. Separate emails with a comma. If left blank, it will include all emails from the exams. Selection occurres before mutation.",
)
@click.option(
    "--blacklist-emails",
    default=None,
    help="This is a list of emails you want the autograder to skip. Separate emails with a comma. If left blank, it will not blacklist any email. Blacklist occurres before mutation.",
)
@click.option(
    "--mutate-emails",
    default=None,
    help="This is a json dictionary which maps the email on examtool to the default email on gradescope ({str:str}). It will not mutate emails which are not in the list. If this is left blank, it will not mutate any emails.",
)
@click.option(
    "--question-numbers",
    default=None,
    help="This is a list of question numbers you want to autograde to the assignment (Numbers are defined by the Gradescope question number). Separate question numbers with a comma. If left blank, it will grade all questions from the exams.",
)
@click.option(
    "--blacklist-question-numbers",
    default=None,
    help="This is a list of question numbers you want the autograder to skip (Numbers are defined by the Gradescope question number). Separate question numbers with a comma. If left blank, it will not blacklist any questions.",
)
@click.option(
    "--create/--update",
    default=True,
    help="Create will generate the outline and set the grouping type, update will ",
)
@click.option(
    "--custom-grouper",
    default=None,
    help="This is the path to a python file which contains the dictionary named EXACTLY `examtool_custom_grouper_fns` mapping question IDs or question Gradescope numbers to a function which returns a QuestionGrouper type. See examtool.api.gradescope_autograde for details about that function.",
)
@click.option(
    "--jobs",
    "-j",
    default=10,
    type=int,
    help="This is the number of simultaneous questions currently being processed. Default: 10",
)
@click.option(
    "--sub-jobs",
    "-sj",
    default=10,
    type=int,
    help="This is the number of simultaneous jobs of a question currently being processed. Note this is per question. Default: 10",
)
@hidden_target_folder_option
def gradescope_autograde(
    exam,
    name_question,
    sid_question,
    course,
    assignment,
    assignment_title,
    email,
    password,
    token,
    emails,
    blacklist_emails,
    mutate_emails,
    question_numbers,
    blacklist_question_numbers,
    create,
    custom_grouper,
    jobs,
    sub_jobs,
    target,
):
    """
    Uploads and autogrades the given exam(s).
    """
    exam = [e.strip() for e in exam.split(",")]
    target = target or "out/export/" + exam[0]

    grouper_map = None
    if custom_grouper:
        if os.path.exists(custom_grouper):
            import importlib.util

            spec = importlib.util.spec_from_file_location("module.name", custom_grouper)
            cg = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(cg)
            grouper_map = cg.examtool_custom_grouper_fns

    grader = GradescopeGrader(
        email=email,
        password=password,
        gs_login_tokens_path=token,
        simultaneous_jobs=jobs,
        simultaneous_sub_jobs=sub_jobs,
    )

    email_mutation_list = None
    if mutate_emails:
        with open(mutate_emails, "r") as f:
            email_mutation_list = json.load(f)

    def extract_list(s):
        return [i.strip() for i in s.split(",")]

    if emails is not None:
        emails = extract_list(emails)
    if blacklist_emails is not None:
        blacklist_emails = extract_list(blacklist_emails)
    if question_numbers is not None:
        question_numbers = extract_list(question_numbers)
    if blacklist_question_numbers is not None:
        blacklist_question_numbers = extract_list(blacklist_question_numbers)
    if create or assignment is None:
        grader.main(
            exam,
            target,
            name_question,
            sid_question,
            course,
            gs_assignment_id=assignment,
            gs_assignment_title=assignment_title,
            emails=emails,
            blacklist_emails=blacklist_emails,
            email_mutation_list=email_mutation_list,
            question_numbers=question_numbers,
            blacklist_question_numbers=blacklist_question_numbers,
            custom_grouper_map=grouper_map,
        )
    else:
        grader.add_additional_exams(
            exam,
            target,
            name_question,
            sid_question,
            course,
            assignment,
            emails=emails,
            blacklist_emails=blacklist_emails,
            email_mutation_list=email_mutation_list,
            question_numbers=question_numbers,
            blacklist_question_numbers=blacklist_question_numbers,
            custom_grouper_map=grouper_map,
        )


if __name__ == "__main__":
    gradescope_autograde()



import csv
from json import loads

import click
from cryptography.fernet import Fernet

from examtool.api.utils import rand_id
from examtool.api.database import process_ok_exam_upload, set_exam, get_exam, set_roster
from examtool.api.extract_questions import extract_questions, get_name
from examtool.api.scramble import is_compressible_group
from examtool.cli.utils import exam_name_option, verify_roster


@click.command()
@exam_name_option
@click.option(
    "--json",
    prompt=True,
    type=click.File("r"),
    help="The exam JSON you wish to deploy.",
)
@click.option(
    "--roster",
    prompt=True,
    type=click.File("r"),
    help="The roster CSV you wish to deploy.",
)
@click.option(
    "--start-time",
    prompt=True,
    type=int,
    help="The unix timestamp corresponding to the start time of the exam.",
)
@click.option(
    "--enable-clarifications",
    prompt=True,
    default=True,
    type=bool,
    help="Let students send clarifications to staff from the exam itself.",
)
def deploy(exam, json, roster, start_time, enable_clarifications):
    """
    Deploy an exam to the website. You must specify an exam JSON and associated roster CSV.
    You can deploy the JSON multiple times and the password will remain unchanged.
    """
    json = json.read()
    roster = csv.reader(roster, delimiter=",")

    exam_content = loads(json)

    exam_content["default_deadline"] = 0
    exam_content["secret"] = Fernet.generate_key().decode("utf-8")

    try:
        old_secret = get_exam(exam=exam)["secret"]
        if old_secret:
            print("Reusing old secret...")
            exam_content["secret"] = old_secret
    except:
        pass

    set_exam(exam=exam, json=exam_content)
    roster = list(roster)
    if not verify_roster(roster=roster):
        return
    roster = roster[1:]  # ditch headers
    set_roster(exam=exam, roster=roster)

    print("Exam uploaded with password:", exam_content["secret"][:-1])

    print("Exam deployed to https://exam.cs61a.org/{}".format(exam))

    print("Initializing announcements...")
    elements = list(extract_questions(exam_content, include_groups=True))
    for element in elements:
        element["id"] = element.get("id", rand_id())  # add IDs to groups
    elements = {
        element["id"]: get_name(element)
        for element in elements
        if element["type"] != "group" or not is_compressible_group(element)
    }

    students = [
        {
            "email": email,
            "start_time": start_time,
            "end_time": int(deadline),
            "no_watermark": bool(int(rest[0]) if rest else False),
        }
        for email, deadline, *rest in roster
    ]

    print("Updating announcements roster with {} students...".format(len(students)))

    process_ok_exam_upload(
        exam=exam,
        data={
            "students": students,
            "questions": [
                {"canonical_question_name": name} for name in elements.values()
            ],
        },
        clear=True,
        enable_clarifications=enable_clarifications,
    )

    print("Announcements deployed to https://announcements.cs61a.org")


if __name__ == "__main__":
    deploy()



from __future__ import annotations

import random
from dataclasses import dataclass
from examtool.api.utils import as_list

from examtool.gui_files.svg import SVGGraphic


@dataclass(frozen=True)
class Point:
    x: int
    y: int

    def __iter__(self):
        yield self.x
        yield self.y

    def dist(self, other: Point):
        return (self.x - other.x) ** 2 + (self.y - other.y) ** 2


@as_list
def get_watermark_points(seed):
    old_seed = int(random.random() * 100000)
    random.seed(seed)
    for _ in range(20):
        yield Point(random.randrange(100), random.randrange(100))
    random.seed(old_seed)


def create_watermark(seed, *, brightness, scale=2):
    graphic = SVGGraphic(100 * scale, 100 * scale)
    color = f"rgb({255 - brightness}, {255 - brightness / 2}, 0)"
    graphic.draw_line(0, 0, 10, 10, color)
    graphic.draw_line(0, 10, 10, 0, color)
    for x, y in get_watermark_points(seed):
        graphic.draw_rect(x * scale, y * scale, scale, scale, color, color)
    return str(graphic)



import json
import re
import os

import pypandoc
from tqdm import tqdm
from multiprocessing.pool import ThreadPool
from os.path import dirname

from examtool.api.utils import list_to_dict, IDFactory

VERSION = 2  # increment when backward-incompatible changes are made


def html_convert(x):
    return pypandoc.convert_text(x, "html5", "md", ["--mathjax"])


def tex_convert(x):
    return pypandoc.convert_text(x, "latex", "md")


class LineBuffer:
    def __init__(self, text, *, src_map=None):
        self.lines = text.strip().split("\n")
        self.src_map = src_map
        self.i = 0

    def _pop(self) -> str:
        if self.i == len(self.lines):
            raise SyntaxError("File terminated unexpectedly")
        self.i += 1
        return self.lines[self.i - 1]

    def pop(self) -> str:
        line = self._pop()
        stripped = line.rstrip()
        while stripped.endswith("\\"):
            if stripped.endswith(r"\\"):
                line = stripped[:-1]
                break
            line = stripped[:-1] + "\n" + self._pop()
            stripped = line.rstrip()
        return line

    def empty(self):
        return self.i == len(self.lines)

    def location(self):
        if self.src_map is None:
            return [self.i, "<string>"]
        else:
            return self.src_map[self.i - 1]


def parse_directive(line):
    if not any(
        line.startswith(f"# {x} ")
        for x in ["BEGIN", "END", "INPUT", "CONFIG", "DEFINE", "IMPORT"]
    ):
        return None, None, None
    tokens = line.split(" ", 3)
    return (
        tokens[1],
        tokens[2] if len(tokens) > 2 else "",
        tokens[3] if len(tokens) > 3 else "",
    )


def process_title(line):
    tokens = line.split(" ")
    point_sec = tokens[-1]
    has_fixed = tokens and tokens[0] == "FIXED"
    if has_fixed:
        tokens = tokens[1:]
        line = " ".join(tokens)
    if not point_sec or point_sec[0] != "[" or point_sec[-1] != "]":
        return line.strip(), has_fixed, None
    try:
        return " ".join(tokens[:-1]).strip(), has_fixed, float(point_sec[1:-1])
    except ValueError:
        return line.strip(), has_fixed, None


class ToParse:
    def __init__(self, text, type):
        self.text = text
        self.type = type

        self.html = None
        self.tex = None


def parse(text):
    return {"text": text, "html": ToParse(text, "html"), "tex": ToParse(text, "tex")}


def parse_define(directive, rest, defines):
    defines["substitutions"] = defines.get("substitutions", {})
    defines["substitutions_match"] = defines.get("substitutions_match", [])
    defines["substitution_groups"] = defines.get("substitution_groups", [])
    defines["substitution_ranges"] = defines.get("substitution_ranges", {})

    if directive == "MATCH":
        regex = r"\[(.*)\]\s+\[(.*)\]"
        matches = re.match(regex, rest)
        if not matches or len(matches.groups()) != 2:
            raise SyntaxError("Invalid declaration of DEFINE MATCH")
        directives, replacements = matches.groups()
        directives_list = directives.split(" ")
        replacements_list = replacements.split(" ")
        if len(replacements_list) < len(directives_list):
            raise SyntaxError(
                "DEFINE MATCH must have at least as many replacements as it has directives"
            )
        defines["substitutions_match"].append(
            {"directives": directives_list, "replacements": replacements_list}
        )
    elif directive == "GROUP":
        regex = r"(\([^()]+\)\s*)+"
        matches = re.match(regex, rest)
        if not matches:
            raise SyntaxError("Invalid declaration of DEFINE GROUP")
        blocks = re.findall(r"\([^()]+\)", rest)
        if len(blocks) <= 1:
            raise SyntaxError("DEFINE GROUP is incomplete")
        for i, block in enumerate(blocks):
            blocks[i] = tuple(block[1:-1].split(" "))
        if not all(len(block) == len(blocks[0]) for block in blocks):
            raise SyntaxError("DEFINE GROUP blocks must all be of the same length")
        defines["substitution_groups"].append(
            {
                "directives": blocks[0],
                "replacements": list_to_dict(
                    [list_to_dict(block) for block in blocks[1:]]
                ),
            }
        )
    elif directive == "RANGE":
        blocks = rest.split(" ")
        if len(blocks) != 3:
            raise SyntaxError("DEFINE RANGE takes exactly three arguments")
        directive, low, high = blocks
        try:
            low = int(low)
            high = int(high)
        except ValueError:
            raise SyntaxError("DEFINE RANGE bounds must be integers")
        defines["substitution_ranges"][directive] = [low, high]
    else:
        defines["substitutions"][directive] = rest.split(" ")


def parse_input_lines(lines):
    if not lines:
        raise SyntaxError("No INPUT directives found in QUESTION")
    _, directive, rest = parse_directive(lines[0])
    correct_options = []
    if directive == "OPTION" or directive == "SELECT":
        options = []
        existing_options = set()
        for line in lines:
            _, other_directive, rest = parse_directive(line)
            if other_directive != directive:
                raise SyntaxError("Multiple INPUT types found in a single QUESTION")
            fixed = "FIXED "
            is_fixed = False
            if rest.startswith(fixed):
                is_fixed = True
                rest = rest[len(fixed) :]

            correct = "CORRECT "
            is_correct = False
            if rest.startswith(correct):
                is_correct = True
                rest = rest[len(correct) :]

            if is_correct:
                correct_options.append(rest)

            if rest in existing_options:
                raise SyntaxError("Cannot have duplicate INPUT options")
            existing_options.add(rest)

            options.append(parse(rest))
            options[-1]["fixed"] = is_fixed
        return (
            "multiple_choice" if directive == "OPTION" else "select_all",
            options,
            correct_options,
        )
    elif directive in (
        "SHORT_ANSWER",
        "SHORT_CODE_ANSWER",
        "LONG_ANSWER",
        "LONG_CODE_ANSWER",
    ):
        if len(lines) > 1:
            raise SyntaxError(
                "Multiple INPUT directives found for a {}".format(directive)
            )
        if directive == "SHORT_ANSWER":
            return "short_answer", None, None
        elif directive == "SHORT_CODE_ANSWER":
            return "short_code_answer", None, None
        try:
            num_lines = int(rest or "10")
        except TypeError:
            raise SyntaxError("Expected integer as option for {}".format(directive))
        if directive == "LONG_ANSWER":
            return "long_answer", num_lines, None
        elif directive == "LONG_CODE_ANSWER":
            return "long_code_answer", num_lines, None
    raise SyntaxError("Unrecognized directive: {}".format(directive))


def consume_rest_of_solution(buff, end):
    out = []
    while True:
        line = buff.pop()
        mode, directive, rest = parse_directive(line)
        if mode is None:
            out.append(line)
        elif mode == "END":
            if directive == end:
                return parse("\n".join(out))
            else:
                raise SyntaxError(
                    f"Unexpected END ({directive if directive else line}) in SOLUTION"
                )
        else:
            raise SyntaxError(
                f"Unexpected directive ({mode if mode else line}) in SOLUTION"
            )


def consume_rest_of_question(buff, id_factory):
    contents = []
    input_lines = []
    defines = {}
    solution = None
    solution_note = None
    config = {}
    template = []
    while True:
        line = buff.pop()
        mode, directive, rest = parse_directive(line)
        if mode is None:
            if input_lines and line.strip():
                template.append(line)
            elif not input_lines:
                contents.append(line)
        elif mode == "INPUT":
            input_lines.append(line)
        elif mode == "BEGIN":
            if directive == "SOLUTION":
                solution = consume_rest_of_solution(buff, directive)
            elif directive == "NOTE":
                solution_note = consume_rest_of_solution(buff, directive)
            else:
                raise SyntaxError(
                    f"Unexpected BEGIN ({directive if directive else line}) in QUESTION"
                )
        elif mode == "END":
            if directive == "QUESTION":
                (
                    question_type,
                    options,
                    option_solutions,
                ) = parse_input_lines(input_lines)

                if option_solutions and solution:
                    raise SyntaxError("Received multiple solutions.")

                template = "\n".join(template)
                if template:
                    if question_type in ("long_answer", "long_code_answer"):
                        # OK
                        pass
                    elif question_type in ("short_answer", "short_code_answer"):
                        if "\n" in template:
                            raise SyntaxError(
                                "Cannot have newlines in template for INPUT SHORT ANSWER"
                            )
                    else:
                        raise SyntaxError(
                            f"Cannot have a template for question type {question_type}"
                        )

                return {
                    "id": id_factory.get_id(config.get("ID")),
                    "type": question_type,
                    "solution": {
                        "solution": solution,
                        "options": option_solutions,
                        "note": solution_note,
                    },
                    **parse("\n".join(contents)),
                    "config": config,
                    "options": options,
                    "template": template,
                    **defines,
                }
            else:
                raise SyntaxError(
                    f"Unexpected END {directive if directive else line} in QUESTION"
                )
        elif mode == "DEFINE":
            parse_define(directive, rest, defines)
        elif mode == "CONFIG":
            config[directive] = rest
        else:
            raise SyntaxError(
                f"Unexpected directive ({mode if mode else line}) in QUESTION"
            )


def consume_rest_of_group(buff, end, id_factory):
    group_contents = []
    elements = []
    started_elements = False
    defines = {}
    pick_some = None
    scramble = False
    inline = False
    while True:
        line = buff.pop()
        mode, directive, rest = parse_directive(line)
        if mode is None:
            if started_elements and line.strip():
                raise SyntaxError("Unexpected text in GROUP after QUESTIONs started")
            elif not started_elements:
                group_contents.append(line)
        elif mode == "BEGIN" and directive == "QUESTION":
            started_elements = True
            title, is_fixed, points = process_title(rest)
            if title:
                raise SyntaxError(
                    "Unexpected arguments passed in BEGIN QUESTION directive"
                )
            question = consume_rest_of_question(buff, id_factory)
            question["points"] = points
            question["fixed"] = is_fixed
            elements.append(question)
        elif mode == "BEGIN" and directive == "GROUP":
            started_elements = True
            title, is_fixed, points = process_title(rest)
            group = consume_rest_of_group(buff, "GROUP", id_factory)
            if (title or points) and group["inline"]:
                raise SyntaxError("Cannot create an inline group with title or points")
            group["name"] = title
            group["points"] = points
            group["fixed"] = is_fixed
            elements.append(group)
        elif mode == "END" and directive == end:
            return {
                "type": "group",
                **parse("\n".join(group_contents)),
                "elements": elements,
                **defines,
                "pick_some": pick_some,
                "scramble": scramble,
                "inline": inline,
            }
        elif mode == "DEFINE":
            parse_define(directive, rest, defines)
        elif mode == "CONFIG":
            if directive == "PICK":
                if pick_some:
                    raise SyntaxError("Multiple CONFIG PICK found in GROUP")
                try:
                    pick_some = int(rest)
                except ValueError:
                    raise SyntaxError("Invalid argument passed to CONFIG PICK")
            elif directive == "SCRAMBLE":
                scramble = True
            elif directive == "INLINE":
                inline = True
            else:
                raise SyntaxError(
                    f"Unexpected CONFIG directive ({directive if directive else line}) in GROUP"
                )
        else:
            raise SyntaxError(f"Unexpected directive ({line}) in GROUP")


def _convert(text, *, path=None, allow_random_ids=True):
    groups = []
    public = None
    config = {}
    defines = {}
    watermark = None
    if path is not None:
        buff = load_imports(text, path)
    else:
        buff = LineBuffer(text)
    id_factory = IDFactory(allow_random_ids=allow_random_ids)
    try:
        while not buff.empty():
            line = buff.pop()
            if not line.strip():
                continue
            mode, directive, rest = parse_directive(line)
            if mode == "CONFIG":
                if directive in [
                    "SCRAMBLE_GROUPS",
                    "SCRAMBLE_QUESTIONS",
                    "SCRAMBLE_OPTIONS",
                ]:
                    config[directive.lower()] = [int(x) for x in rest.split(" ") if x]
                elif directive == "WATERMARK":
                    watermark = dict(brightness=int(rest))
                else:
                    raise SyntaxError(
                        "Unexpected CONFIG directive {}".format(directive)
                    )
            elif mode == "BEGIN" and directive in ["GROUP", "PUBLIC"]:
                title, is_fixed, points = process_title(rest)
                group = consume_rest_of_group(buff, directive, id_factory)
                group["name"] = title
                group["points"] = points
                group["fixed"] = is_fixed
                if (title.strip() or points) and group["inline"]:
                    raise SyntaxError(
                        "Cannot create an inline group with a title or points"
                    )
                if directive == "PUBLIC":
                    if public:
                        raise SyntaxError("Only one PUBLIC block is allowed")
                    if is_fixed:
                        raise SyntaxError("PUBLIC blocks are already FIXED")
                    public = group
                else:
                    groups.append(group)
            elif mode == "DEFINE":
                parse_define(directive, rest, defines)
            else:
                raise SyntaxError(f"Unexpected directive: {line}")
    except SyntaxError as e:
        line_num, file = buff.location()
        raise SyntaxError(
            "Parse stopped on {}:{} with error: {}".format(file, line_num, e)
        )

    return {
        "public": public,
        "groups": groups,
        "config": config,
        **defines,
        "watermark": watermark,
        "version": VERSION,
    }


def pandoc(target, *, draft=False, num_threads):
    to_parse = []

    def explore(pos):
        if isinstance(pos, ToParse):
            to_parse.append(pos)
        elif isinstance(pos, dict):
            for child in pos.values():
                explore(child)
        elif isinstance(pos, list):
            for child in pos:
                explore(child)

    explore(target)

    pandoc_delimiter = """\n\nDELIMITER\n\n"""

    if draft:

        def transpile_target(t):
            return pandoc_delimiter.join(x.text for x in to_parse if x.type == t)

        html = html_convert(transpile_target("html")).split(
            html_convert(pandoc_delimiter)
        )
        tex = tex_convert(transpile_target("tex")).split(tex_convert(pandoc_delimiter))

        for x, h in zip(filter(lambda x: x.type == "html", to_parse), html):
            x.html = h

        for x, t in zip(filter(lambda x: x.type == "tex", to_parse), tex):
            x.tex = t
    else:

        def pandoc_convert(x):
            x.__dict__[x.type] = (
                html_convert(x.text) if x.type == "html" else tex_convert(x.text)
            )

        with ThreadPool(num_threads) as p:
            list(
                tqdm(
                    p.imap_unordered(pandoc_convert, to_parse),
                    total=len(to_parse),
                    desc="Parts Processed",
                    unit="Part",
                )
            )

    def pandoc_dump(obj):
        assert isinstance(obj, ToParse)
        return obj.__dict__[obj.type]

    return json.dumps(target, default=pandoc_dump)


def convert(text, *, path=None, draft=False, allow_random_ids=True, num_threads):
    return json.loads(
        convert_str(
            text,
            path=path,
            draft=draft,
            allow_random_ids=allow_random_ids,
            num_threads=num_threads,
        )
    )


def convert_str(
    text,
    *,
    path=None,
    draft=False,
    allow_random_ids=True,
    num_threads=16,
):
    return pandoc(
        _convert(text, path=path, allow_random_ids=allow_random_ids),
        draft=draft,
        num_threads=num_threads,
    )


def import_file(filepath: str) -> str:
    if not filepath:
        raise SyntaxError("IMPORT must take in a filepath")
    with open(filepath, "r") as f:
        return f.read()


def load_imports(base_text: str, base_path: str):
    lines = []

    def _load(text: str, path: str):
        for i, line in enumerate(text.split("\n")):
            mode, directive, rest = parse_directive(line)
            if mode == "IMPORT":
                filepath = os.path.join(
                    dirname(path), " ".join([directive, rest]).rstrip()
                )
                try:
                    _load(import_file(filepath), filepath)
                except FileNotFoundError:
                    raise SyntaxError(
                        f"Parse stopped on {path}:{i + 1}: Unable to import {filepath}"
                    )
            else:
                lines.append([i + 1, path, line])

    _load(base_text, base_path)

    line_strs = []
    src_map = []
    for line_num, path, line in lines:
        line_strs.append(line)
        src_map.append([line_num, path])

    return LineBuffer("\n".join(line_strs), src_map=src_map)



"""
Developed by ThaumicMekanism [Stephan K.] - all credit goes to him!
"""
import contextlib
import sys
from typing import Callable, List

from multiprocessing.pool import ThreadPool
from tqdm.contrib import DummyTqdmFile

import examtool.api.download
from examtool.api.gradescope_upload import APIClient
from examtool.api.extract_questions import (
    extract_groups,
    extract_questions,
    extract_public,
)
from fullGSapi.api.login_tokens import LoginTokens
from fullGSapi.api.client import GradescopeClient
from fullGSapi.api.assignment_grader import (
    GS_Crop_info,
    GS_Outline,
    GS_assignment_Grader,
    GS_Outline_Question,
    GS_Question,
    GroupTypes,
    RubricItem,
    QuestionRubric,
)
import os
import time
from tqdm import tqdm

def_tqdm_args = {"dynamic_ncols": True}


@contextlib.contextmanager
def std_out_err_redirect_tqdm():
    orig_out_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = map(DummyTqdmFile, orig_out_err)
        yield orig_out_err[0]
    # Relay exceptions
    except Exception as exc:
        raise exc
    # Always restore sys.stdout/err if necessary
    finally:
        sys.stdout, sys.stderr = orig_out_err


class GradescopeGrader:
    def __init__(
        self,
        email: str = None,
        password: str = None,
        gs_login_tokens: LoginTokens = None,
        gs_login_tokens_path: str = None,
        simultaneous_jobs: int = 10,
        simultaneous_sub_jobs: int = 10,
    ):
        print(f"Setting up the Gradescope Grader...")
        entered_email_pwd = email is not None and password is not None

        logged_in = False

        if gs_login_tokens is None and not entered_email_pwd:
            gs_login_tokens = LoginTokens.load(gs_login_tokens_path)
            if gs_login_tokens is not None:
                logged_in = True
        else:
            print(
                "Ignoring current token file since you entered an email and password.\nLogging in with those credentials..."
            )
        if gs_login_tokens is None:
            gs_login_tokens = LoginTokens(path=gs_login_tokens_path)

        if entered_email_pwd:
            logged_in = gs_login_tokens.login(email, password)

        if not logged_in:
            gs_login_tokens.prompt_login(email=email)

        self.gs_login_tokens = gs_login_tokens
        self.gs_client = gs_login_tokens.gsFullapi
        self.gs_api_client = gs_login_tokens.gsAPI

        self.simultaneous_jobs = simultaneous_jobs
        self.simultaneous_sub_jobs = simultaneous_sub_jobs
        print(f"Finished setting up the Gradescope Grader")

    def main(
        self,
        exams: [str],
        out: str,
        name_question_id: str,
        sid_question_id: str,
        gs_class_id: str,
        gs_assignment_id: str = None,  # If none, we will create a class.
        gs_assignment_title: str = "Examtool Exam",
        emails: [str] = None,
        blacklist_emails: [str] = None,
        email_mutation_list: {str: str} = {},
        question_numbers: [str] = None,
        blacklist_question_numbers: [str] = None,
        custom_grouper_map: {
            str: Callable[[str, GS_Question, dict, dict], "QuestionGrouper"]
        } = None,
    ):
        if gs_assignment_title is None:
            gs_assignment_title = "Examtool Exam"
        if not exams:
            raise ValueError(
                "You must specify at least one exam you would like to upload!"
            )

        out = out or "out/export/" + exams[0]

        (
            exam_json,
            email_to_data_map,
            question_page_mapping,
        ) = self.fetch_and_export_examtool_exam_data(
            exams,
            out,
            name_question_id,
            sid_question_id,
            emails=emails,
            email_mutation_list=email_mutation_list,
        )

        # Remove blacklisted emails
        if blacklist_emails is not None:
            for bemail in blacklist_emails:
                email_to_data_map.pop(bemail, None)

        # Create assignment if one is not already created.
        if gs_assignment_id is None:
            print("Creating the gradescope assignment...")
            outline_path = f"{out}/OUTLINE.pdf"
            gs_assignment_id = self.create_assignment(
                gs_class_id, gs_assignment_title, outline_path
            )
            if not gs_assignment_id:
                raise ValueError(
                    "Did not receive a valid assignment id. Did assignment creation fail?"
                )
            print(f"Created gradescope assignment with id {gs_assignment_id}!")
        else:
            print(f"Using assignment ({gs_assignment_id}) which was already created!")

        # Lets now get the assignment grader
        grader: GS_assignment_Grader = self.get_assignment_grader(
            gs_class_id, gs_assignment_id
        )

        # Now that we have the assignment and outline pdf, lets generate the outline.
        print("Generating the examtool outline...")
        examtool_outline = ExamtoolOutline(
            grader,
            exam_json,
            [name_question_id, sid_question_id],
            question_page_mapping,
        )

        # Finally we need to upload and sync the outline.
        print("Uploading the generated outline...")
        self.upload_outline(grader, examtool_outline)

        # We can now upload the student submission since we have an outline
        print("Uploading student submissions...")
        failed_uploads = self.upload_student_submissions(
            out, gs_class_id, gs_assignment_id, emails=email_to_data_map.keys()
        )

        # Removing emails which failed to upload
        if failed_uploads:
            print(
                f"Removing emails which failed to upload. Note: These will NOT be graded! {failed_uploads}"
            )
            for email in tqdm(failed_uploads, **def_tqdm_args):
                email_to_data_map.pop(email)

        # For each question, group, add rubric and grade
        print("Setting the grade type for grouping for each question...")
        gs_outline = examtool_outline.get_gs_outline()
        self.set_group_types(gs_outline)

        # Fetch the student email to question id map
        print("Fetching the student email to submission id's mapping...")
        email_to_question_sub_id = grader.email_to_qids()

        # Check to see which emails may not be in the Gradescope roster and attempt to correct
        self.attempt_fix_unknown_gs_email(
            email_to_question_sub_id,
            email_to_data_map,
            name_question_id=name_question_id,
            sid_question_id=sid_question_id,
        )

        # Finally we can process each question
        self.process_questions(
            gs_outline,
            question_numbers,
            blacklist_question_numbers,
            email_to_data_map,
            email_to_question_sub_id,
            name_question_id,
            sid_question_id,
            custom_grouper_map,
        )

    def process_questions(
        self,
        gs_outline,
        question_numbers,
        blacklist_question_numbers,
        email_to_data_map,
        email_to_question_sub_id,
        name_question_id,
        sid_question_id,
        custom_grouper_map,
    ):
        def proc_q(d):
            qid, question = d
            if (
                question_numbers is not None
                and qid not in question_numbers
                or blacklist_question_numbers is not None
                and qid in blacklist_question_numbers
            ):
                tqdm.write(f"[{qid}]: Skipping!")
                return
            tqdm.write(f"[{qid}]: Processing question...")
            try:
                self.process_question(
                    qid,
                    question.get_gs_question(),
                    email_to_data_map,
                    email_to_question_sub_id,
                    name_question_id,
                    sid_question_id,
                    custom_grouper_map,
                )
            except Exception as e:
                import traceback

                traceback.print_exc(file=tqdm)
                tqdm.write(str(e))

        qi = list(gs_outline.questions_iterator())
        with ThreadPool(self.simultaneous_jobs) as p:
            list(
                tqdm(
                    p.imap_unordered(proc_q, qi),
                    total=len(qi),
                    desc="Questions Graded",
                    unit="Question",
                    **def_tqdm_args,
                )
            )

    def add_additional_exams(
        self,
        exams: [str],
        out: str,
        name_question_id: str,
        sid_question_id: str,
        gs_class_id: str,
        gs_assignment_id: str,
        emails: [str] = None,
        blacklist_emails: [str] = None,
        email_mutation_list: {str: str} = {},
        question_numbers: [str] = None,
        blacklist_question_numbers: [str] = None,
        custom_grouper_map: {
            str: Callable[[str, GS_Question, dict, dict], "QuestionGrouper"]
        } = None,
    ):
        """
        If emails is None, we will import the entire exam, if it has emails in it, it will only upload submissions
        from the students in the emails list contained in the exams list. If the student has submissions in multiple exams,
        the tool will warn you and ask which exam you would like to use as the student submission.
        """
        if not exams:
            raise ValueError(
                "You must specify at least one exam you would like to upload!"
            )
        if email_mutation_list is None:
            email_mutation_list = {}

        out = out or "out/export/" + exams[0]

        (
            exam_json,
            email_to_data_map,
            question_page_mapping,
        ) = self.fetch_and_export_examtool_exam_data(
            exams,
            out,
            name_question_id,
            sid_question_id,
            emails=emails,
            email_mutation_list=email_mutation_list,
        )

        # Remove blacklisted emails
        if blacklist_emails is not None:
            for bemail in blacklist_emails:
                email_to_data_map.pop(bemail, None)

        # Lets now get the assignment grader
        grader: GS_assignment_Grader = self.get_assignment_grader(
            gs_class_id, gs_assignment_id
        )

        # Now that we have the assignment and outline pdf, lets generate the outline.
        print("Generating the examtool outline...")
        examtool_outline = ExamtoolOutline(
            grader,
            exam_json,
            [name_question_id, sid_question_id],
            question_page_mapping,
        )

        # Merge the outline with the existing one
        outline = grader.get_outline()
        if not outline:
            raise ValueError("Failed to fetch the existing outline")
        examtool_outline.merge_gs_outline_ids(outline)

        # We can now upload the student submission since we have an outline
        print("Uploading student submissions...")
        failed_uploads = self.upload_student_submissions(
            out, gs_class_id, gs_assignment_id, emails=email_to_data_map.keys()
        )

        # Removing emails which failed to upload
        if failed_uploads:
            print(
                f"Removing emails which failed to upload. Note: These will NOT be graded! {failed_uploads}"
            )
            for email in failed_uploads:
                email_to_data_map.pop(email)

        # Fetch the student email to question id map
        print("Fetching the student email to submission id's mapping...")
        email_to_question_sub_id = grader.email_to_qids()

        # Check to see which emails may not be in the Gradescope roster and attempt to correct
        self.attempt_fix_unknown_gs_email(
            email_to_question_sub_id,
            email_to_data_map,
            name_question_id=name_question_id,
            sid_question_id=sid_question_id,
        )

        gs_outline = examtool_outline.get_gs_outline()

        # Finally we can process each question
        self.process_questions(
            gs_outline,
            question_numbers,
            blacklist_question_numbers,
            email_to_data_map,
            email_to_question_sub_id,
            name_question_id,
            sid_question_id,
            custom_grouper_map,
        )

    def fetch_and_export_examtool_exam_data(
        self,
        exams: [str],
        out: str,
        name_question_id: str,
        sid_question_id: str,
        emails: [str] = None,
        email_mutation_list: {str: str} = {},
    ):
        """
        Fetches the submissions from the exams in the exams list.
        If the emails list is None, it will fetch all emails, if it has emails in it, it will only return data for those emails.
        The mutation step occurres after the specific emails selection stage if applicable.
        The mutation list comes in the form of current email to new email.

        Returns:
        exam_json - The json of the exam
        email_to_data_map - the mapping of emails to their data.
        """
        if not exams:
            raise ValueError(
                "You must specify at least one exam you would like to upload!"
            )
        if email_mutation_list is None:
            email_mutation_list = {}

        print("Downloading exams data...")
        exam_json = None
        email_to_data_map = {}
        email_to_exam_map = {}

        first_exam = True
        for exam in exams:
            (
                tmp_exam_json,
                tmp_template_questions,
                tmp_email_to_data_map,
                tmp_total,
            ) = examtool.api.download.download(exam)

            # Choose only the emails we want to keep.
            if emails:
                for email in list(tmp_email_to_data_map.keys()):
                    if email not in emails:
                        tmp_email_to_data_map.pop(email, None)

            # Next, we want to mutate any emails
            for orig_email, new_email in email_mutation_list.items():
                if orig_email not in tmp_email_to_data_map:
                    print(
                        f"WARNING: Could not perform mutation on email {orig_email} (to {new_email}) because it does not exist in the data map!"
                    )
                    continue
                if new_email in tmp_email_to_data_map:
                    print(
                        f"Could not mutate email {new_email} (from {orig_email}) as the original email is already in the data map!"
                    )
                    continue
                tmp_email_to_data_map[new_email] = tmp_email_to_data_map.pop(orig_email)

            # Finally, we should merge together the student responses.
            for email, data in tmp_email_to_data_map.items():
                if email in email_to_data_map:
                    print(
                        f"WARNING: Student with email {email} submitted to multiple exams!"
                    )

                    def prompt_q():
                        input_data = None
                        while not input_data:
                            print(
                                f"Student's current responses are from {email_to_exam_map[email]}, would you like to use {exam} instead?"
                            )
                            input_data = input("[y/n]> ")
                            if input_data.lower() in ["y", "yes"]:
                                return True
                            if input_data.lower() in ["n", "no"]:
                                return False
                            print("Please type yes or no!")

                    if not prompt_q():
                        continue
                email_to_exam_map[email] = exam
                email_to_data_map[email] = data

            question_page_mapping = examtool.api.download.get_question_to_page_mapping(
                tmp_template_questions,
                exam,
                out,
                name_question_id,
                sid_question_id,
            )

            print(f"[{exam}]: Exporting exam pdfs...")

            self.export_exam(
                tmp_template_questions,
                tmp_email_to_data_map,
                tmp_total,
                exam,
                out,
                name_question_id,
                sid_question_id,
                include_outline=first_exam,
            )

            # Set global data for the examtool
            if first_exam:
                first_exam = False
                exam_json = tmp_exam_json

        # Lets finally clean up the student responses
        self.cleanse_student_response_data(email_to_data_map)

        return exam_json, email_to_data_map, question_page_mapping

    def attempt_fix_unknown_gs_email(
        self,
        email_to_question_sub_id,
        email_to_data_map,
        name_question_id,
        sid_question_id,
    ):
        def prompt_fix(old_email, name, sid):
            input_data = None
            while not input_data:
                print(
                    f"Could not find {old_email} (name: {name}; sid: {sid}) in Gradescope! Please enter the Gradescope email of the student or `skip` to remove this student from autograding."
                )
                input_data = input("> ")
                if "@" in input_data.lower():
                    return input_data
                if input_data.lower() in ["n", "no", "skip"]:
                    return False
                print(
                    "The input is not a valid email (you are missing the `@`)! If you would like to skip, type `skip` or `no`."
                )

        remove_email = ["DUMMY"]
        map_email = {}
        while remove_email or map_email:
            remove_email = []
            map_email = {}
            for email, data in email_to_data_map.items():
                if email not in email_to_question_sub_id:
                    responses = data["responses"]
                    name = responses.get(name_question_id, None)
                    sid = responses.get(sid_question_id, None)
                    new_email = prompt_fix(email, name, sid)
                    if new_email:
                        map_email[email] = new_email
                    else:
                        print(
                            f"Skipping {email}! This will remove the email from the data map."
                        )
                        remove_email.append(email)
            for email, new_email in map_email.items():
                email_to_data_map[new_email] = email_to_data_map.pop(email)
            for email in remove_email:
                email_to_data_map.pop(email)

    def cleanse_student_response_data(self, email_to_data_map: dict):
        for email, data in email_to_data_map.items():
            std_questions = data["student_questions"]
            std_responses = data["responses"]
            for question in std_questions:
                qid = question["id"]
                if qid not in std_responses:
                    std_responses[qid] = (
                        []
                        if question["type"] in ["multiple_choice", "select_all"]
                        else ""
                    )

    def export_exam(
        self,
        template_questions,
        email_to_data_map,
        total,
        exam,
        out,
        name_question_id,
        sid_question_id,
        include_outline=True,
    ):
        examtool.api.download.export(
            template_questions,
            email_to_data_map,
            total,
            exam,
            out,
            name_question_id,
            sid_question_id,
            include_outline=include_outline,
        )

    def create_assignment(self, gs_class_id: str, gs_title: str, outline_path: str):
        assignment_id = self.gs_client.create_exam(gs_class_id, gs_title, outline_path)
        if not assignment_id:
            print("Failed to create the exam! Make sure it has a unique title.")
            return
        return assignment_id

    def get_assignment_grader(
        self, gs_class_id: str, assignment_id: str
    ) -> GS_assignment_Grader:
        return self.gs_client.get_assignment_grader(gs_class_id, assignment_id)

    def upload_outline(
        self, grader: GS_assignment_Grader, examtool_outline: "ExamtoolOutline"
    ):
        outline = grader.update_outline(examtool_outline.get_gs_outline())
        if not outline:
            raise ValueError("Failed to upload or get the outline")
        examtool_outline.merge_gs_outline_ids(outline)

    def upload_student_submissions(
        self, out: str, gs_class_id: str, assignment_id: str, emails: [str] = None
    ):
        failed_emails = []
        email_files = []
        for file_name in os.listdir(out):
            if "@" not in file_name:
                continue
            student_email = file_name[:-4]
            if emails and student_email not in emails:
                continue
            email_files.append((file_name, student_email))
        with std_out_err_redirect_tqdm() as orig_stdout:
            # for file_name, student_email in tqdm(
            #     email_files, file=orig_stdout, unit="Submission", **def_tqdm_args
            # ):
            def func(tup):
                file_name, student_email = tup
                if not self.gs_api_client.upload_pdf_submission(
                    gs_class_id,
                    assignment_id,
                    student_email,
                    os.path.join(out, file_name),
                ):
                    failed_emails.append(student_email)

            with ThreadPool(self.simultaneous_jobs) as p:
                list(
                    tqdm(
                        p.imap_unordered(func, email_files),
                        total=len(email_files),
                        file=orig_stdout,
                        unit="Submission",
                        **def_tqdm_args,
                    )
                )
        return failed_emails

    def set_group_types(self, outline: GS_Outline, debug=True):
        questions = list(outline.questions_iterator())
        with std_out_err_redirect_tqdm() as orig_stdout:
            # for qid, question in tqdm(
            #     questions, file=orig_stdout, unit="Question", **def_tqdm_args
            # ):
            #     self.set_group_type(question)
            def sgt(q):
                qid, question = q
                self.set_group_type(question)

            with ThreadPool(self.simultaneous_jobs) as p:
                list(
                    tqdm(
                        p.imap_unordered(sgt, questions),
                        total=len(questions),
                        file=orig_stdout,
                        unit="Question",
                        **def_tqdm_args,
                    )
                )

    def set_group_type(self, o_question: GS_Outline_Question):
        question_type = o_question.data.get("type")
        q = o_question.get_gs_question()
        q_type = GroupTypes.complex
        if question_type in ["select_all", "multiple_choice"]:
            q_type = GroupTypes.mc
        # if question_type in ["long_answer", "long_code_answer"]:
        #     q_type = GroupTypes.non_grouped
        return q.set_group_type(q_type)

    def process_question(
        self,
        qid: str,
        question: GS_Question,
        email_to_data_map: dict,
        email_to_question_sub_id_map: dict,
        name_question_id: str,
        sid_question_id: str,
        custom_grouper_map: {
            str: Callable[[str, GS_Question, dict, dict], "QuestionGrouper"]
        },
    ):
        # Group questions
        if question.data.get("id") in [name_question_id, sid_question_id]:
            tqdm.write("Skipping grouping of an id question!")
            return
        tqdm.write(f"[{qid}]: Grouping...")
        groups = self.group_question(
            qid,
            question,
            email_to_data_map,
            email_to_question_sub_id_map,
            custom_grouper_map,
        )
        if groups:
            # Group answers
            tqdm.write(f"[{qid}]: Syncing groups on gradescope...")
            self.sync_groups_on_gradescope(qid, question, groups)
            tqdm.write(f"[{qid}]: Syncing rubric items...")
            rubric = self.sync_rubric(qid, question, groups)
            # in here, add check to see if qid is equal to either name or sid q id so we do not group those.
            tqdm.write(f"[{qid}]: Applying grades for each group...")
            self.grade_question(qid, question, rubric, groups)
        else:
            tqdm.write(f"[{qid}]: Failed to group question {qid}!")

    def group_question(
        self,
        qid: str,
        question: GS_Question,
        email_to_data_map: dict,
        email_to_question_sub_id_map: dict,
        custom_grouper_map: {
            str: Callable[[str, GS_Question, dict, dict], "QuestionGrouper"]
        },
    ):
        if custom_grouper_map is not None:
            examtool_qid = question.data.get("id")
            if examtool_qid in custom_grouper_map:
                return custom_grouper_map[examtool_qid](
                    qid, question, email_to_data_map, email_to_question_sub_id_map
                )
            if qid in custom_grouper_map:
                return custom_grouper_map[qid](
                    qid, question, email_to_data_map, email_to_question_sub_id_map
                )
        # Default handler
        qtype = question.data.get("type")
        if qtype in ["multiple_choice", "select_all"]:
            return self.group_mc_question(
                qid, question, email_to_data_map, email_to_question_sub_id_map
            )
        elif qtype in ["short_answer", "short_code_answer"]:
            return self.group_short_ans_question(
                qid, question, email_to_data_map, email_to_question_sub_id_map
            )
        elif qtype in ["long_answer", "long_code_answer"]:
            return self.group_long_ans_question(
                qid, question, email_to_data_map, email_to_question_sub_id_map
            )
        else:
            tqdm.write(
                f"Unsupported question type {qtype} for question {question.data}!"
            )
            return None

    def group_mc_question(
        self,
        qid: str,
        question: GS_Question,
        email_to_data_map: dict,
        email_to_question_sub_id_map: dict,
        custom_rubric_weights_fn: Callable[
            [GS_Question, List[str], List[bool]], List[float]
        ] = None,
    ):
        data = question.data
        # This is a list of correct options from left (top) to right (bottom)
        correct_seq = []
        seq_name = []
        solution_options = data.get("solution", {})
        if solution_options is not None:
            solution_options = solution_options.get("options", [])
        if solution_options is None:
            solution_options = []
        all_options = [option.get("text") for option in data.get("options", [])]
        for option in all_options:
            correct_seq.append(option in solution_options)
            seq_name.append(option)

        # Add blank option
        correct_seq.append(None)
        seq_name.append("Blank")
        # Add student did not receive this question
        correct_seq.append(None)
        seq_name.append("Student did not receive this question")

        rubric_weights = (
            self.get_basic_rubric_scores(question, seq_name, correct_seq)
            if custom_rubric_weights_fn is None
            else custom_rubric_weights_fn(question, seq_name, correct_seq)
        )

        groups = QuestionGrouper(
            question,
            rubric=[
                RubricItem(description=item[0], weight=item[1])
                for item in zip(seq_name, rubric_weights)
            ],
        )

        def list_to_str(l):
            s = ""
            for item in l:
                s += str(int(item))
            return s

        eqid = question.data["id"]
        for email, data in email_to_data_map.items():
            responses = data.get("responses", {})
            response = responses.get(eqid)
            selection = [False] * len(correct_seq)
            if response is None:
                selection[-1] = True
            elif response == []:
                selection[-2] = True
            else:
                if not isinstance(response, list):
                    response = [response]
                for i, option in enumerate(all_options):
                    selection[i] = option in response

            s = list_to_str(selection)
            sid = email_to_question_sub_id_map[email][qid]
            if s not in groups:
                groups.add_group(QuestionGroup(s, selection))
            groups.get_group(s).add_sid(sid)
        return groups

    def group_short_ans_question(
        self,
        qid: str,
        question: GS_Question,
        email_to_data_map: dict,
        email_to_question_sub_id_map: dict,
        lower_check: bool = True,
        custom_rubric_weights_fn: Callable[
            [GS_Question, List[str], List[bool]], List[float]
        ] = None,
        strip_md_from_sol: bool = True,
    ):
        data = question.data
        # This is a list of correct options from left (top) to right (bottom)
        solution = data.get("solution", {})
        if solution is not None:
            solution = solution.get("solution", {})
            if solution is not None:
                solution = solution.get("text")
        if not solution:
            tqdm.write(
                f"[{qid}]: No solution defined for this question! Only grouping blank and std did not receive."
            )
            solution = "Correct"
        correct_seq = [True]
        seq_name = [solution]

        # Add a wrong option
        correct_seq.append(None)
        seq_name.append("Incorrect")
        # Add blank option
        correct_seq.append(None)
        seq_name.append("Blank")
        # Add student did not receive this question
        correct_seq.append(None)
        seq_name.append("Student did not receive this question")

        rubric_weights = (
            self.get_basic_rubric_scores(question, seq_name, correct_seq)
            if custom_rubric_weights_fn is None
            else custom_rubric_weights_fn(question, seq_name, correct_seq)
        )

        groups = QuestionGrouper(
            question,
            rubric=[
                RubricItem(description=item[0], weight=item[1])
                for item in zip(seq_name, rubric_weights)
            ],
        )

        # Process solution
        if lower_check:
            sol = solution.strip().lower()
        else:
            sol = solution.strip()

        if strip_md_from_sol:

            def strip_part(text, boundary):
                if text.startswith(boundary) and text.endswith(boundary):
                    blen = len(boundary)
                    return (text[blen:-blen], True)
                else:
                    return (text, False)

            sol, replaced = strip_part(sol, "$")

            if not replaced:
                sol, replaced = strip_part(sol, "```")

                if not replaced:
                    sol, replaced = strip_part(sol, "`")

        eqid = question.data["id"]
        for email, data in email_to_data_map.items():
            responses = data.get("responses", {})
            response = responses.get(eqid)
            selection = [False] * len(correct_seq)
            if response is None:
                selection[-1] = True
                response = "Student did not receive this question"
            elif response == "":
                selection[-2] = True
                response = "Blank"
            else:
                if solution is not None:
                    same = None
                    if lower_check:
                        same = response.lower().strip() == sol
                    else:
                        same = response.strip() == sol
                    if same:
                        selection[0] = True
                    else:
                        selection[1] = True

            sid = email_to_question_sub_id_map[email][qid]
            if response not in groups:
                groups.add_group(QuestionGroup(response, selection))
            groups.get_group(response).add_sid(sid)
        return groups

    def group_long_ans_question(
        self,
        qid: str,
        question: GS_Question,
        email_to_data_map: dict,
        email_to_question_sub_id_map: dict,
    ):
        """
        We will only be grouping students who did not get the question or left it blank.
        """
        data = question.data
        # This is a list of correct options from left (top) to right (bottom)
        correct_seq = [True]
        seq_name = ["Correct"]

        # Add blank option
        correct_seq.append(None)
        seq_name.append("Blank")
        # Add student did not receive this question
        correct_seq.append(None)
        seq_name.append("Student did not receive this question")

        rubric_weights = self.get_long_ans_rubric_scores(
            question, seq_name, correct_seq
        )

        groups = QuestionGrouper(
            question,
            rubric=[
                RubricItem(description=item[0], weight=item[1])
                for item in zip(seq_name, rubric_weights)
            ],
        )

        group_blank = QuestionGroup("Blank", [False, True, False])
        groups.add_group(group_blank)

        group_sdnrtq = QuestionGroup(
            "Student did not receive this question", [False, False, True]
        )
        groups.add_group(group_sdnrtq)

        eqid = question.data["id"]
        for email, data in email_to_data_map.items():
            responses = data.get("responses", {})
            response = responses.get(eqid)
            if not response:
                sid = email_to_question_sub_id_map[email][qid]
                if response is None:
                    group_sdnrtq.add_sid(sid)
                elif response == "":
                    group_blank.add_sid(sid)
        return groups

    def sync_groups_on_gradescope(
        self, qid: str, question: GS_Question, groups: "QuestionGrouper"
    ):
        """
        Groups is a list of name, submission_id, selected answers
        """
        failed_groups_names = []
        i = 1
        failed = False
        while not question.is_grouping_ready():
            timeout = 5
            tqdm.write(
                f"[{qid}]: Question grouping not ready! Retrying in {timeout} seconds!"
            )
            time.sleep(timeout)
        #     print(f"[{qid}]: Question grouping not ready! Retrying in {timeout} seconds" + (" " * timeout), end="\r")
        #     for i in range (timeout):
        #         print(f"[{qid}]: Question grouping not ready! Retrying in {timeout} seconds" + ("." * (1 + i)), end="\r")
        #         time.sleep(1)
        #     failed = True
        # if failed:
        #     print("")
        gradescope_groups = question.get_groups()

        def all_zeros(s: str):
            return s and all(v == "0" for v in s)

        def set_group(group, gs_group):
            group.set_id(gs_group.get("id"))

        for group in groups.get_groups():
            g_name = group.get_name()
            for gs_group in gradescope_groups:
                if gs_group["question_type"] == "mc":
                    # The question type is mc so lets group by the internal mc
                    if g_name == "Blank":
                        # This is the blank group, lets use the internal label to group
                        if all_zeros(gs_group["internal_title"]):
                            set_group(group, gs_group)
                    else:
                        flip_g_name = g_name[:-2][::-1]
                        if gs_group["internal_title"] is not None:
                            if (
                                flip_g_name == gs_group["internal_title"]
                                and g_name[len(g_name) - 1] != "1"
                            ):
                                set_group(group, gs_group)
                        else:
                            if g_name == gs_group["title"]:
                                set_group(group, gs_group)
                else:
                    # The question type is not mc so we should group on title and internal title for blank.
                    # The internal title should only say Blank for default blank grouped submissions.
                    # We then check the normal title if this is not true
                    if (
                        g_name == gs_group["internal_title"]
                        or g_name == gs_group["title"]
                    ):
                        set_group(group, gs_group)

        def submit_group(group, question, failed_groups_names, max_attempts):
            attempt = 1
            g_name = group.get_name()
            sids = group.get_sids()
            if not sids:
                # We do not want to create groups which no questions exist.
                return
            group_id = group.get_id()
            while attempt < max_attempts:
                if not group_id:
                    group_id = question.add_group(g_name)
                if group_id is None:
                    attempt += 1
                    time.sleep(1)
                    return
                if not question.group_submissions(group_id, sids):
                    tqdm.write(
                        f"[{qid}]: Failed to group submissions to {group_id}. SIDS: {sids}"
                    )
                    failed_groups_names.append(g_name)
                break
            else:
                tqdm.write(f"[{qid}]: Failed to create group for {g_name}! ({groups})")
                failed_groups_names.append(g_name)

        # max_attempts = 5
        # for group in tqdm(
        #     groups.get_groups(),
        #     desc=f"[{qid}]: Syncing Groups",
        #     unit="Group",
        #     **def_tqdm_args,
        # ):
        #     submit_group(group, question, failed_groups_names, max_attempts)

        gps = groups.get_groups()

        def sg(g):
            submit_group(g, question, failed_groups_names, 5)

        with ThreadPool(self.simultaneous_sub_jobs) as p:
            list(
                tqdm(
                    p.imap_unordered(sg, gps),
                    total=len(gps),
                    desc=f"[{qid}]: Syncing Groups",
                    unit="Group",
                    **def_tqdm_args,
                )
            )

        # This is to decrease down stream errors
        for failed_group_name in failed_groups_names:
            groups.remove(failed_group_name)

    @classmethod
    def get_basic_rubric_scores(cls, question: GS_Question, group_names, correct_seq):
        scores = []
        num_correct = sum([1 for correct in correct_seq if correct])
        num_choices = sum([1 for correct in correct_seq if correct is not None])
        points = question.data.get("points", 1)
        if points is None:
            points = 1
        rubric_weight = 0
        if num_correct != 0:
            rubric_weight = (1 / num_correct) * points
        for correct in correct_seq:
            if correct is None:
                scores.append(0)
            else:
                if correct:
                    scores.append(rubric_weight)
                else:
                    scores.append(-rubric_weight)
        return scores

    @classmethod
    def get_long_ans_rubric_scores(
        cls, question: GS_Question, group_names, correct_seq
    ):
        return [0] * len(correct_seq)

    def sync_rubric(
        self, qid: str, question: GS_Question, groups: "QuestionGrouper"
    ) -> QuestionRubric:
        rubric = QuestionRubric(question)
        if len(groups) == 0:
            return rubric

        qrubric: [RubricItem] = groups.get_rubric()

        if len(rubric) == 1:
            default_rubric_item = rubric.get_rubric_items()[0]
            if default_rubric_item.description == "Correct":
                first_item = qrubric[0]
                if not rubric.update_rubric_item(
                    default_rubric_item,
                    description=first_item.description,
                    weight=first_item.weight,
                ):
                    tqdm.write(
                        f'[{qid}]: Failed to update default "Correct" rubric item!'
                    )
                # qrubric.remove(first_item)

        existing_rubric_items = rubric.get_rubric_items()
        existing_rubric_items_desc = [
            item.description for item in existing_rubric_items
        ]

        for rubric_item in tqdm(
            qrubric, desc=f"[{qid}]: Syncing Rubric", unit="Rubric", **def_tqdm_args
        ):
            if rubric_item.description not in existing_rubric_items_desc:
                rubric.add_rubric_item(rubric_item)

        return rubric

    def grade_question(
        self, qid: str, question: GS_Question, rubric: QuestionRubric, groups: dict
    ):
        question_data = question.get_question_info()
        sub_id_mapping = {str(sub["id"]): sub for sub in question_data["submissions"]}
        # for group in tqdm(
        #     groups.get_groups(), desc=f"[{qid}]: Grading", unit="Group", **def_tqdm_args
        # ):
        def sg(group):
            group_sel = group.get_selected_items()
            group_sids = group.get_sids()
            if len(group_sids) > 0:
                sid = group_sids[0]
                if not sub_id_mapping[str(sid)]["graded"]:
                    if not rubric.grade(sid, group_sel, save_group=True):
                        tqdm.write(
                            f"[{qid}]: Failed to grade group {group.get_name()}!"
                        )

        gps = groups.get_groups()
        with ThreadPool(self.simultaneous_sub_jobs) as p:
            list(
                tqdm(
                    p.imap_unordered(sg, gps),
                    total=len(gps),
                    desc=f"[{qid}]: Grading",
                    unit="Group",
                    **def_tqdm_args,
                )
            )


class ExamtoolOutline:
    name_region = GS_Crop_info(1, 2.4, 11.4, 99, 18.8)
    sid_region = GS_Crop_info(1, 2.4, 18.9, 99, 28.7)

    def __init__(
        self,
        grader: GS_assignment_Grader,
        exam_json: dict,
        id_question_ids: List[str],
        question_page_mapping: List[int],
    ):
        self.exam_json = exam_json
        self.gs_number_to_exam_q, self.gs_outline = self.generate_gs_outline(
            grader, exam_json, id_question_ids, question_page_mapping
        )

    def get_gs_crop_info(self, page, question=None):
        return GS_Crop_info(page, 4, 4, 96, 96)

    def question_to_gso_question(
        self, grader: GS_assignment_Grader, page, question: dict
    ) -> GS_Outline_Question:
        weight = question.get("points")
        if not weight:
            weight = 0
        return GS_Outline_Question(
            grader,
            None,
            [self.get_gs_crop_info(page, question=question)],
            title=question.get("name", ""),
            weight=weight,
        )

    def generate_gs_outline(
        self,
        grader: GS_assignment_Grader,
        exam_json: dict,
        id_question_ids: [str],
        question_page_mapping: List[int],
    ):
        gs_number_to_exam_q = {}
        questions = []

        page = 0  # Page 1 is an info page

        qid = 1
        if exam_json.get("public"):
            prev_page = 0
            pg = GS_Outline_Question(
                grader,
                None,
                [self.get_gs_crop_info(page, exam_json.get("public"))],
                title="Public",
                weight=0,
            )
            sqid = 1
            for question in extract_public(exam_json):
                question_id = question.get("id")
                if question_id in id_question_ids:
                    print(f"Skipping {question_id} as it is an id question.")
                    page += 1  # Still need to increment this as it is still on the exam pdf.
                    continue
                pg.add_child(
                    self.question_to_gso_question(
                        grader, question_page_mapping[page], question
                    )
                )
                gs_number_to_exam_q[f"{qid}.{sqid}"] = question
                sqid += 1
                page += 1
            if page != prev_page and len(pg.children) > 0:
                questions.append(pg)
                qid += 1

        for group in extract_groups(exam_json):
            prev_page = page
            weight = group.get("points", "0")
            if not weight:
                weight = 0
            g = GS_Outline_Question(
                grader,
                None,
                [self.get_gs_crop_info(question_page_mapping[page], group)],
                title=group.get("name", ""),
                weight=weight,
            )
            sqid = 1
            for question in extract_questions(
                group, extract_public_bool=False, top_level=False
            ):
                g.add_child(
                    self.question_to_gso_question(
                        grader, question_page_mapping[page], question
                    )
                )
                gs_number_to_exam_q[f"{qid}.{sqid}"] = question
                sqid += 1
                page += 1
            if page != prev_page:
                questions.append(g)
                qid += 1

        outline = GS_Outline(self.name_region, self.sid_region, questions)
        return (gs_number_to_exam_q, outline)

    def get_gs_outline(self):
        return self.gs_outline

    def merge_gs_outline_ids(self, outline: GS_Outline):
        self.gs_outline = outline
        for qnum, q in outline.questions_iterator():
            q.data = self.gs_number_to_exam_q[qnum]

    def questions_iterator(self):
        yield from self.gs_outline.questions_iterator()


class QuestionGroup:
    def __init__(self, name: str, selected_rubric_items: [bool], gid: str = None):
        self.name = name
        self.selected_rubric_items = (
            selected_rubric_items  # Bool array of selected items.
        )
        self.gid = gid
        self.sids = set()

    def get_name(self):
        return self.name

    def get_id(self):
        return self.gid

    def set_id(self, gid: str):
        self.gid = gid

    def get_sids(self):
        return list(self.sids)

    def add_sid(self, sid: str):
        self.sids.add(sid)

    def add_sids(self, sids: [str]):
        self.sids = self.sids.union(sids)

    def get_selected_items(self):
        return self.selected_rubric_items


class QuestionGrouper:
    def __init__(
        self,
        question: GS_Question,
        rubric: [RubricItem],  # This is a list of rubric items.
        groups: {str: QuestionGroup} = None,
    ):
        self.groups = groups
        if not self.groups:
            self.groups = {}
        self.question = question
        self.rubric = rubric

    def get_groups(self):
        return self.groups.values()

    def get_group(self, name):
        return self.groups.get(name)

    def add_group(self, group: QuestionGroup):
        self.groups[group.get_name()] = group

    def remove(self, group_name):
        for g in self.groups:
            if g.get_name() == group_name:
                self.groups.remove(g)
                return

    def __len__(self):
        return len(self.groups)

    def get_rubric(self) -> [RubricItem]:
        return self.rubric

    def __contains__(self, key):
        return key in self.groups



import csv
import os
import pathlib
from multiprocessing.pool import ThreadPool

import click
from tqdm import tqdm

from examtool.api.render_html_export import render_html_exam
from examtool.api.render_pdf_export import render_pdf_exam
from examtool.cli.utils import exam_name_option, hidden_output_folder_option
import examtool.api.download
import examtool.api.assemble_export


@click.command()
@exam_name_option
@click.option(
    "--name-question",
    default=None,
    help="The ID of the question for the student's name.",
)
@click.option(
    "--sid-question", default=None, help="The ID of the question for the student's SID."
)
@click.option(
    "--with-substitutions/--without-substitutions",
    default=False,
    help="Include keyword substitutions in exported question bodies.",
)
@click.option(
    "--via-html/--direct-pdf",
    default=False,
    help="Use an HTML-based PDF exporter or export directly to PDF.",
)
@click.option(
    "--num-threads",
    default=16,
    type=int,
    help="The number of threads to process the JSON file.",
)
@hidden_output_folder_option
def download(
    exam, out, name_question, sid_question, with_substitutions, via_html, num_threads
):
    """
    Download student submissions for an exam.
    Exams are downloaded as PDFs into a target folder - specify `out` to redirect the folder.
    An `OUTLINE.pdf` is also generated for Gradescope, as is a `summary.csv` for analytics or autograding.
    """
    out = out or "out/export/" + exam
    pathlib.Path(out).mkdir(parents=True, exist_ok=True)

    (
        exam_json,
        template_questions,
        email_to_data_map,
        total,
    ) = examtool.api.download.download(exam)

    with open(os.path.join(out, "summary.csv"), "w") as f:
        writer = csv.writer(f)
        for row in total:
            writer.writerow(row)

    assembled_exams = examtool.api.assemble_export.export(
        template_questions,
        email_to_data_map,
        exam,
        name_question,
        sid_question,
        substitute_in_question_text=with_substitutions,
    )

    def render(name_exam):
        name, exam = name_exam

        target = os.path.join(out, f"{name}.pdf")

        if via_html:
            export = render_html_exam(exam)
            export(target)

        else:
            pdf = render_pdf_exam(exam)
            pdf.output(target)

    with ThreadPool(num_threads) as p:
        list(
            tqdm(
                p.imap_unordered(render, assembled_exams.items()),
                total=len(assembled_exams),
                desc="Rendering",
                unit="Exam",
            )
        )


if __name__ == "__main__":
    download()



from examtool.api.scramble import is_compressible_group


def extract_questions(
    exam,
    extract_public_bool: bool = True,
    top_level: bool = True,
    include_groups: bool = False,
    nest_all: bool = False,
):
    def merge_text(parent, child):
        for attr in ["text", "html", "tex"]:
            child[attr] = parent["name"] + "\n\n" + parent[attr] + "\n\n" + child[attr]

    def group_questions(group):
        out = _group_questions(group)
        try:
            first = next(out)
        except StopIteration:
            return
        merge_text(group, first)
        yield first
        for child in out:
            if nest_all:
                merge_text(group, child)
            yield child

    def _group_questions(group):
        for i, element in enumerate(
            group.get("elements", []) + group.get("questions", [])
        ):
            element["index"] = f"{group.get('index', '0.')}{i + 1}."
            if element.get("type") == "group":
                if include_groups:
                    yield element
                out = group_questions(element)
                yield from out
            else:
                yield element

    if extract_public_bool:
        yield from extract_public(exam)
    if top_level:
        for i, group in enumerate(exam["groups"]):
            group["index"] = str(i + 1) + "."
            if include_groups:
                yield group
            yield from group_questions(group)
    else:
        yield from group_questions(exam)


def extract_public(exam):
    if exam.get("public"):
        yield from extract_questions(
            exam["public"], extract_public_bool=False, top_level=False
        )


def extract_groups(group):
    for g in group["groups"]:
        if is_compressible_group(g):
            for g2 in g["elements"]:
                yield g2
        else:
            yield g


def get_name(element):
    if "name" in element:
        return f"{element['index']} {element['name']}"
    else:
        return element["index"]



import re

import click

from datetime import date, datetime
from collections import OrderedDict

exam_name_option = click.option(
    "--exam", prompt=True, default="cs61a-test-final", help="The exam name."
)
hidden_output_folder_option = click.option(
    "--out",
    default=None,
    help="Output folder. Leave as default for dependent commands to work.",
    type=click.Path(),
)
hidden_target_folder_option = click.option(
    "--target",
    default=None,
    help="Target folder for PDFs. Leave as default unless the source output folder is not the default.",
    type=click.Path(),
)


def verify_roster(*, roster):
    for i, row in enumerate(roster):
        j = len(row)
        if j != 2 and j != 3:
            print(
                f"ValueError: The roster must contain 2 or 3 columns: Email, Timestamp, Skip Watermarks [Optional]. "
                f"Found {j} item(s) on row {i + 1}: {row}"
            )
            return False
    return True


def prettify(course_code):
    m = re.match(r"([a-z]+)([0-9]+[a-z]?)", course_code)
    return m and (m.group(1) + " " + m.group(2)).upper()


def determine_semester():
    sem_map = OrderedDict()
    sem_map["1/1"] = "Spring"
    sem_map["6/1"] = "Summer"
    sem_map["9/1"] = "Fall"
    today = date.today()

    def compare_date(DATE_TO_COMPARE):
        parsed = (
            datetime.strptime(DATE_TO_COMPARE, "%m/%d").date().replace(year=today.year)
        )
        return parsed > today

    sem = next(iter(sem_map))
    for DATE_TO_COMPARE, new_sem in sem_map.items():
        if compare_date(DATE_TO_COMPARE):
            break
        sem = new_sem
    return f"{sem} {today.year}"



import csv
import json
import os
from collections import defaultdict
from dataclasses import asdict

import click

from examtool.api.database import get_roster
from examtool.api.substitution_finder import find_unexpected_words
from examtool.cli.utils import exam_name_option, hidden_target_folder_option


@click.command()
@exam_name_option
@hidden_target_folder_option
@click.option(
    "--out",
    type=click.File("w"),
    default=None,
    help="Output a CSV containing the list of cheaters",
)
def cheaters(exam, target, out):
    """
    Identify potential instances of cheating.
    """
    if not target:
        target = "out/logs/" + exam
    logs = defaultdict(list)
    for email, deadline in get_roster(exam=exam):
        short_target = os.path.join(target, email)
        full_target = os.path.join(target, "full", email)
        if os.path.exists(short_target):
            with open(short_target) as f:
                logs[email].extend(json.load(f))
        if os.path.exists(full_target):
            with open(full_target) as f:
                data = json.load(f)
                for record in data:
                    if "snapshot" not in record:
                        print(email, record)
                    else:
                        logs[email].append(
                            {**record["snapshot"], "timestamp": record["timestamp"]}
                        )
                        logs[email].append(
                            {**record["history"], "timestamp": record["timestamp"]}
                        )
    logs = list(logs.items())
    suspects = find_unexpected_words(exam, logs)
    if out:
        if suspects:
            writer = csv.writer(out)
            keys = asdict(suspects[0])
            writer.writerow(list(keys))
            for suspect in suspects:
                writer.writerow(asdict(suspect).values())
                suspect.explain()
    else:
        for suspect in suspects:
            suspect.explain()



import os
import random
import string
from functools import wraps

from typing import Optional


def as_list(func):
    @wraps(func)
    def wrapped(*args, **kwargs):
        return list(func(*args, **kwargs))

    return wrapped


def sanitize_email(email):
    return email.replace("_", r"\_")


def dict_to_list(d):
    out = [None] * len(d)
    for k, v in d.items():
        out[int(k)] = v
    return out


def rand_id(length=32):
    return "".join(random.choices(string.ascii_uppercase, k=length))


def list_to_dict(l):
    return {i: x for i, x in enumerate(l)}


def rel_path(path):
    root = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(root, path)


def rel_open(path, *args, **kwargs):
    return open(rel_path(path), *args, **kwargs)


class IDFactory:
    def __init__(self, *, id_start="", id_end="", length=32, allow_random_ids=True):
        if length is None:
            length = 32
        self.length = length
        self.id_start = id_start
        self.id_end = id_end
        self.allow_random_ids = allow_random_ids
        self.current_ids = set()

    def id_from_str(self, string):
        return self.id_start + string + self.id_end

    def get_id(self, string: Optional[str] = None):
        if string is None:
            if not self.allow_random_ids:
                raise SyntaxError(
                    "A custom ID is required but was not set for this question!"
                )
            qid = rand_id(length=self.length)
            while qid in self.current_ids:
                qid = rand_id(length=self.length)
        else:
            qid = self.id_from_str(string)
            if qid in self.current_ids:
                raise SyntaxError(f"Received duplicate question IDs {qid}.")

        self.current_ids.add(qid)
        return qid



class SVGRect:
    def __init__(self, x, y, width, height, stroke, fill):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.stroke = stroke
        self.fill = fill

    def __str__(self):
        return """<rect x="{0}" y="{1}" width="{2}" height="{3}" stroke="{4}" fill="{5}" />""".format(
            self.x, self.y, self.width, self.height, self.stroke, self.fill
        )


class SVGCircle:
    def __init__(self, x, y, radius, stroke, fill):
        self.x = x
        self.y = y
        self.radius = radius
        self.stroke = stroke
        self.fill = fill

    def __str__(self):
        return (
            """<circle cx="{0}" cy="{1}" r="{2}" stroke="{3}" fill="{4}" />""".format(
                self.x, self.y, self.radius, self.stroke, self.fill
            )
        )


class SVGLine:
    def __init__(self, x1, y1, x2, y2, stroke):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.stroke = stroke

    def __str__(self):
        return """<line x1="{0}" y1="{1}" x2="{2}" y2="{3}" stroke="{4}" />""".format(
            self.x1, self.y1, self.x2, self.y2, self.stroke
        )


class SVGPolygon:
    def __init__(self, points, stroke, fill):
        self.points = points  # list of lists
        self.stroke = stroke
        self.fill = fill

    def __str__(self):
        points_str = " ".join(",".join(map(str, point)) for point in self.points)
        return """<polygon points="{0}" stroke="{1}" fill="{2}"/>""".format(
            points_str, self.stroke, self.fill
        )


class SVGText:
    def __init__(self, x, y, text, stroke, fill, font_size, font_family):
        self.x = x
        self.y = y
        self.text = text
        self.stroke = stroke
        self.fill = fill
        self.font_size = font_size
        self.font_family = font_family

    def __str__(self):
        return """<text x="{0}" y="{1}" stroke="{2}" fill="{3}" font-size="{4}" font-family="{5}">{6}</text>""".format(
            self.x,
            self.y,
            self.stroke,
            self.fill,
            self.font_size,
            self.font_family,
            self.text,
        )


class SVGGraphic:
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.shapes = []

    def draw_rect(self, x, y, width, height, stroke, fill):
        self.shapes.append(SVGRect(x, y, width, height, stroke, fill))

    def draw_circle(self, x, y, radius, stroke, fill):
        self.shapes.append(SVGCircle(x, y, radius, stroke, fill))

    def draw_line(self, x1, y1, x2, y2, stroke):
        self.shapes.append(SVGLine(x1, y1, x2, y2, stroke))

    def draw_polygon(self, points, stroke, fill):
        self.shapes.append(SVGPolygon(points, stroke, fill))

    def write_text(self, x, y, text, stroke, fill, font_size, font_family):
        self.shapes.append(SVGText(x, y, text, stroke, fill, font_size, font_family))

    def __str__(self):
        shapes = "".join(str(shape) for shape in self.shapes)
        return """<svg width="{0}" height="{1}" xmlns="http://www.w3.org/2000/svg">{2}</svg>""".format(
            self.width, self.height, shapes
        )


def create_graphic(width, height):
    return SVGGraphic(width, height)


def draw_rect(graphic, x, y, width, height, stroke="black", fill="black"):
    graphic.draw_rect(x, y, width, height, stroke, fill)


def draw_circle(graphic, x, y, radius, stroke="black", fill="black"):
    graphic.draw_circle(x, y, radius, stroke, fill)


def draw_line(graphic, x1, y1, x2, y2, stroke="black"):
    graphic.draw_line(x1, y1, x2, y2, stroke)


def draw_triangle(graphic, x1, y1, x2, y2, x3, y3, stroke="black", fill="black"):
    graphic.draw_polygon([[x1, y1], [x2, y2], [x3, y3]], stroke, fill)


def write_text(
    graphic,
    x,
    y,
    text,
    stroke="black",
    fill="black",
    font_size="medium",
    font_family="serif",
):
    graphic.write_text(x, y, text, stroke, fill, font_size, font_family)



from os import getenv

from examtool.api.server_delegate import server_only
from examtool.api.utils import as_list

if getenv("ENV") == "SERVER":
    from examtool_web_common.safe_firestore import SafeFirestore
    from google.cloud.exceptions import NotFound

BATCH_SIZE = 400
assert BATCH_SIZE < 500


def clear_collection(db, ref):
    batch = db.batch()
    cnt = 0
    for document in ref.stream():
        batch.delete(document.reference)
        cnt += 1
        if cnt > BATCH_SIZE:
            batch.commit()
            batch = db.batch()
            cnt = 0
    batch.commit()


@server_only
def get_exam(*, exam):
    try:
        db = SafeFirestore()
        out = db.collection("exams").document(exam).get().to_dict()
        if "secret" in out and isinstance(out["secret"], bytes):
            out["secret"] = out["secret"].decode("utf-8")
        return out
    except NotFound:
        raise KeyError


@server_only
def set_exam(*, exam, json):
    db = SafeFirestore()
    db.collection("exams").document(exam).set(json)

    ref = db.collection("exams").document("all")
    data = ref.get().to_dict()
    if exam not in data["exam-list"]:
        data["exam-list"].append(exam)
    ref.set(data)


@server_only
@as_list
def get_roster(*, exam, include_no_watermark=False):
    db = SafeFirestore()
    for student in (
        db.collection("roster").document(exam).collection("deadline").stream()
    ):
        if include_no_watermark:
            yield (
                student.id,
                student.to_dict()["deadline"],
                student.to_dict().get("no_watermark", False),
            )
        else:
            yield student.id, student.to_dict()["deadline"]


@server_only
def set_roster(*, exam, roster):
    db = SafeFirestore()

    ref = db.collection("roster").document(exam).collection("deadline")
    emails = []

    batch = db.batch()
    cnt = 0
    for document in ref.stream():
        batch.delete(document.reference)
        cnt += 1
        if cnt > 400:
            batch.commit()
            batch = db.batch()
            cnt = 0
    batch.commit()

    batch = db.batch()
    cnt = 0
    for email, deadline, *rest in roster:
        assert len(rest) <= 1
        doc_ref = ref.document(email)
        batch.set(
            doc_ref,
            {
                "deadline": int(deadline),
                "no_watermark": bool(int(rest[0]) if rest else False),
            },
        )
        emails.append(email)
        cnt += 1
        if cnt > 400:
            batch.commit()
            batch = db.batch()
            cnt = 0
    batch.commit()

    ref = db.collection("roster").document(exam)
    data = {"all_students": emails}
    ref.set(data)


@server_only
@as_list
def get_submissions(*, exam):
    db = SafeFirestore()

    for ref in db.collection(exam).stream():
        yield ref.id, ref.to_dict()


@server_only
@as_list
def get_logs(*, exam, email):
    db = SafeFirestore()

    for ref in db.collection(exam).document(email).collection("log").stream():
        yield ref.to_dict()


@server_only
@as_list
def get_full_logs(*, exam, email):
    db = SafeFirestore()

    for ref in db.collection(exam).document(email).collection("history").stream():
        yield ref.to_dict()


@server_only
def process_ok_exam_upload(*, exam, data, enable_clarifications=False, clear=True):
    """
    data: {
        "students": [
            {
                "email": string,
                "questions": [
                    {
                        "student_question_name": string,
                        "canonical_question_name": string,
                        "start_time": int,
                        "end_time": int,
                    }
                ],
                "start_time": int,
                "end_time": int,
            }
        ]
        "questions": [
            {
                "canonical_question_name": string,
            }
        ],
    }
    """
    db = SafeFirestore()

    db.collection("exam-alerts").document(exam).set(
        {"questions": data["questions"], "enable_clarifications": enable_clarifications}
    )
    ref = db.collection("exam-alerts").document(exam).collection("students")
    if clear:
        clear_collection(db, ref)

    batch = db.batch()
    cnt = 0
    for student in data["students"]:
        doc_ref = ref.document(student["email"])
        batch.set(doc_ref, student)
        cnt += 1
        if cnt > BATCH_SIZE:
            batch.commit()
            batch = db.batch()
            cnt = 0
    batch.commit()

    ref = db.collection("exam-alerts").document("all")
    exam_list_data = ref.get().to_dict()
    if exam not in exam_list_data["exam-list"]:
        exam_list_data["exam-list"].append(exam)
    ref.set(exam_list_data)



from examtool.api.extract_questions import extract_questions


def get_question_substitutions(original_questions, scrambled_questions, question_id):
    original_question = original_questions[question_id]
    substitutions = {}
    for original, replacement in (
        scrambled_questions[question_id].get("substitutions", {}).items()
    ):
        if original in original_question["text"]:
            substitutions[original] = replacement
    return substitutions


def get_all_substitutions(original_exam, scrambled_exam):
    assert (
        original_exam is not scrambled_exam
    ), "You must make a copy of the original before scrambling it"
    original_questions = {q["id"]: q for q in extract_questions(original_exam)}
    scrambled_questions = {q["id"]: q for q in extract_questions(scrambled_exam)}
    question_substitutions = {}
    for question_id in scrambled_questions:
        question_substitutions[question_id] = get_question_substitutions(
            original_questions, scrambled_questions, question_id
        )
    return question_substitutions



def boolstring(bool):
    return "Matches solution." if bool else "May not match solution."


def grade(email, question, responses, dispatch=None):
    if dispatch:
        if dispatch(email, question):
            return dispatch(email, question)(responses)
    response = responses.get(question["id"])
    if "solution" not in question:
        return "Instant autograder unavailable."

    if response is None:
        return boolstring(False)

    if question["type"] == "multiple_choice":
        return boolstring(response in question["solution"]["options"])
    elif question["type"] == "select_all":
        if question["solution"].get("options"):
            return boolstring(
                sorted(response) == sorted(question["solution"]["options"])
            )
    else:
        if question["solution"].get("solution"):
            return boolstring(response == question["solution"]["solution"]["text"])

    return "Instant autograder unavailable."



import click

from examtool.api.substitution_finder import find_keyword
from examtool.cli.utils import exam_name_option


@click.command()
@exam_name_option
@click.option(
    "--keyword",
    prompt=True,
    help="The keyword you wish to identify.",
)
def identify_keyword(exam, keyword):
    """
    Identify the student from a keyword present in their exam.
    """
    find_keyword(exam, keyword)



import os
import re
import subprocess
from collections import defaultdict
from contextlib import contextmanager

from examtool.api.scramble import latex_escape
from examtool.api.watermarks import create_watermark
from examtool.api.utils import rel_open


def generate(exam, *, include_watermark):
    out = []

    def write(x, *, newline=True):
        out.append(x)
        if newline:
            out.append("\n")

    def write_group(group, is_public):
        if is_public:
            if group["points"] is not None:
                write(rf"{{\bf\item ({group['points']} points)\quad}}")
            else:
                write(r"\item[]", newline=False)
        else:
            if group["points"] is not None:
                write(fr"\q{{{group['points']}}}")
            else:
                write(r"\item", newline=False)
        write(r"{ \bf " + latex_escape(group["name"]) + "}")
        write("\n")
        write(group["tex"])
        write(r"\begin{enumerate}[font=\bfseries]")
        for element in group["elements"]:
            if element["type"] == "group":
                write_group(element, False)
            else:
                write_question(element)
        write(r"\end{enumerate}")
        write(r"\clearpage")

    def write_question(question):
        write(r"\filbreak")
        if question["points"] is not None:
            write(fr"\subq{{{question['points']}}}")
        else:
            write(r"\item \, \hspace{-1em} \, ")
        write(question["tex"])
        solution = question.get("solution", defaultdict())
        if solution.get("solution"):
            write(r"\setlength{\fboxsep}{1em}")
            if "verbatim" in solution["solution"]["tex"]:
                write(
                    r"\catcode`_ 12\relax\n\begin{Verbatim}[frame=single,formatcom=\color{blue},rulecolor=\color{black},xleftmargin=1em,xrightmargin=1em]\n"
                )
                write(
                    solution["solution"]["tex"]
                    .replace("\\begin{verbatim}", "")
                    .replace("\\end{verbatim}", "")
                    .strip()
                )
                write(r"\end{Verbatim}\n\catcode`_ 8\relax")
            else:
                write(r"\fbox{\parbox{0.8\textwidth}{\solution{")
                write(solution["solution"]["tex"])
                write(r"}}}")
        elif question["type"] in ["short_answer", "short_code_answer"]:
            write(r"\framebox[0.8\textwidth][c]{\parbox[c][30px]{0.5\textwidth}{}}")
        elif question["type"] in ["long_answer", "long_code_answer"]:
            write(
                rf"\framebox[0.8\textwidth][c]{{\parbox[c][{30 * (question['options'])}px]{{0.5\textwidth}}{{}}}}"
            )

        correct = solution.get("options", [])

        if question["type"] in ["select_all"]:
            write(r"\begin{options}")
            for option in question["options"]:
                if option["text"] in correct:
                    write(r"\option[\moqs] " + option["tex"])
                else:
                    write(r"\option[\moqb] " + option["tex"])
            write(r"\end{options}")
        if question["type"] in ["multiple_choice"]:
            write(r"\begin{choices}")
            for option in question["options"]:
                if option["text"] in correct:
                    write(r"\option[\mcqs] " + option["tex"])
                else:
                    write(r"\option[\mcqb] " + option["tex"])
            write(r"\end{choices}")
        if solution.get("note"):
            write(r"\\\\ \note{")
            write(solution["note"]["tex"])
            write(r"}")
        write(r"\vspace{10px}")

    if include_watermark:
        write(r"\let\Watermarks=1")

    with rel_open("tex/prefix.tex") as f:
        write(f.read())
    for i, group in enumerate(
        ([exam["public"]] if exam["public"] else []) + exam["groups"]
    ):
        is_public = bool(i == 0 and exam["public"])
        write_group(group, is_public)

    with rel_open("tex/suffix.tex") as f:
        write(f.read())

    return "".join(out)


@contextmanager
def render_latex(exam, subs=None, *, do_twice=False):
    include_watermark = exam.get("watermark") and "value" in exam["watermark"]

    latex = generate(exam, include_watermark=include_watermark)
    latex = re.sub(
        r"\\includegraphics(\[.*\])?{(http.*/(.+))}",
        r"\\immediate\\write18{wget -N \2}\n\\includegraphics\1{\3}",
        latex,
    )
    if subs:
        for k, v in subs.items():
            latex = latex.replace(f"<{k.upper()}>", v)
    if not os.path.exists("temp"):
        os.mkdir("temp")
    with open("temp/out.tex", "w+") as f:
        f.write(latex)

    if include_watermark:
        with open("temp/watermark.svg", "w+") as f:
            f.write(
                create_watermark(
                    exam["watermark"]["value"],
                    brightness=exam["watermark"]["brightness"],
                )
            )
        subprocess.run(
            "inkscape -D -z --file=temp/watermark.svg --export-pdf=temp/watermark.pdf",
            shell=True,
        ).check_returncode()

    subprocess.run(
        "cd temp && pdflatex --shell-escape -interaction=nonstopmode out.tex",
        shell=True,
    )
    if do_twice:
        subprocess.run(
            "cd temp && pdflatex --shell-escape -interaction=nonstopmode out.tex",
            shell=True,
        )
    with open("temp/out.pdf", "rb") as f:
        yield f.read()
    # shutil.rmtree("temp")



from contextlib import contextmanager
from html import escape

import pdfkit

from examtool.api.assemble_export import AssembledExam, OptionQuestion, TextQuestion
from examtool.api.utils import rel_path


def render_html_exam(assembled_exam: AssembledExam):
    blocks = []

    def out(text):
        blocks.append(text.replace("“", '"').replace("”", '"').replace("’", "'"))

    def user_out(text):
        out(escape(text))

    def tag(name):
        @contextmanager
        def with_tag(**attrs):
            out(f"<{name}")
            for k, v in attrs.items():
                if k == "className":
                    k = "class"
                if not v:
                    continue

                out(" ")
                out(k)
                out("=")
                out(repr(str(v)))
            out(">")
            yield
            out(f"</{name}>")

        return with_tag

    h3 = tag("h3")

    div = tag("div")
    label = tag("label")
    input = tag("input")
    pre = tag("pre")

    for question in assembled_exam.questions:
        with div(className="question"):
            with h3():
                out(question.name)

            with div(className="questionText"):
                out(question.prompt.html)

            with div(className="gradable"):
                with div(className="answer"):
                    if isinstance(question, OptionQuestion):
                        for option in question.options:
                            with div(className="checkbox"), label(), input(
                                type="checkbox", checked=option in question.selected
                            ):
                                out(option.html)

                    elif isinstance(question, TextQuestion):
                        with div(className="response"):
                            with pre(style=f"height: {question.height * 3}em"):
                                user_out(question.response)
                    else:
                        assert False, f"Unknown question type {type(question)}"

                with pre(className="autogradeOutput"):
                    user_out(question.autograde_output)

    def export(target):
        pdfkit.from_string(
            "".join(blocks),
            target,
            options={
                "log-level": "none",
                "disable-smart-shrinking": None,
                "zoom": 0.75,
            },
            css=rel_path("css/style.css"),
        )

    return export



from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()


setup(
    name="examtool",
    version="2.2.25b380.dev0",
    author="Rahul Arya",
    author_email="rahularya@berkeley.edu",
    long_description=readme,
    long_description_content_type="text/markdown",
    licence="MIT",
    packages=find_packages(
        include=["examtool.api", "examtool.cli", "examtool.gui_files"]
    ),
    package_data={"": ["**/*.tex", "**/*.css"]},
    include_package_data=True,
    entry_points={"console_scripts": ["examtool=examtool.cli.__main__:cli"]},
    python_requires=">=3.7",
    install_requires=["cryptography"],
    extras_require={
        "admin": [
            "pytz",
            "requests",
            "pypandoc",
            "google-cloud-firestore",
            "google-auth",
            "sendgrid",
        ],
        "cli": [
            "click",
            "pikepdf",
            "pytz",
            "requests",
            "fpdf",
            "pypandoc",
            "sendgrid",
            "func-timeout",
            "fullGSapi>=1.3.7",
            "tqdm",
            "numpy",
            "opencv-python",
            "pdfkit",
        ],
    },
)



import click
import cv2

from examtool.api.database import get_exam, get_roster
from examtool.api.watermark_decoder import decode_watermark
from examtool.api.watermarks import Point
from examtool.cli.utils import exam_name_option


@click.command()
@exam_name_option
@click.option(
    "--image",
    prompt=True,
    type=click.Path(exists=True),
    help="The image or screenshot you wish to identify.",
)
def identify_watermark(exam, image):
    """
    Identify the student from a screenshot containing a watermark.
    """
    img = cv2.imread(image)
    img = cv2.copyMakeBorder(img, 100, 100, 100, 100, cv2.BORDER_CONSTANT)

    corners = []
    bits = []

    def handle_click(event, x, y, flags, params):
        if event == cv2.EVENT_LBUTTONDOWN:
            bits.append(Point(x, y))
            cv2.circle(img, (x, y), 5, (255, 0, 0), -1)
        if event == cv2.EVENT_RBUTTONDOWN:
            corners.append(Point(x, y))
            cv2.circle(img, (x, y), 5, (0, 255, 0), -1)

    cv2.namedWindow("image")
    cv2.setMouseCallback("image", handle_click)
    while True:
        cv2.imshow("image", img)
        if cv2.waitKey(20) & 0xFF == 13:
            break

    print(decode_watermark(get_exam(exam=exam), get_roster(exam=exam), corners, bits))



import http.server
import logging
from os.path import abspath

import requests
import sys
import time
import webbrowser
from urllib.parse import parse_qsl, urlencode, urlparse

log = logging.getLogger(__name__)

CLIENT_ID = "local-dev-email"
CLIENT_SECRET = "KH0mvknMUWT5w3U7zvz6wsUQZoy6UmQ"
OAUTH_SCOPE = "email"

# Localhost
REDIRECT_HOST = "127.0.0.1"
REDIRECT_PORT = 6265

# OAuth post timeout
TIMEOUT = 10

# Server/API config
OK_SERVER_URL = "https://okpy.org"
INFO_ENDPOINT = "/api/v3/user/"
ASSIGNMENT_ENDPOINT = "/api/v3/assignment/"
AUTH_ENDPOINT = "/oauth/authorize"
TOKEN_ENDPOINT = "/oauth/token"
ERROR_ENDPOINT = "/oauth/errors"

# URL to redirect user to upon OAuth success
SUCCESS_ENDPOINT_URL = "https://okpy.org"  # temporary

TOKEN_PATH = ".token"


class OKException(Exception):
    """Base exception class for Bacon/OK integration."""


class OAuthException(OKException):
    """OAuth related exception"""

    def __init__(self, error="", error_description=""):
        super().__init__()
        self.error = error
        self.error_description = error_description


def _pick_free_port(hostname=REDIRECT_HOST, port=0):
    """Try to bind a port. Default=0 selects a free port."""
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind((hostname, port))  # port=0 finds an open port
    except OSError as e:
        if port == 0:
            print("Unable to find an open port for authentication.")
            raise OKException(e)
        else:
            return _pick_free_port(hostname, 0)
    addr, port = s.getsockname()
    s.close()
    return port


def _make_token_post(server, data):
    """Try getting an access token from the server. If successful, returns the
    JSON response. If unsuccessful, raises an OAuthException.
    """
    try:
        response = requests.post(server + TOKEN_ENDPOINT, data=data, timeout=TIMEOUT)
        body = response.json()
    except Exception as e:
        log.warning("Other error when exchanging code", exc_info=True)
        raise OAuthException(error="Authentication Failed", error_description=str(e))
    if "error" in body:
        log.error(body)
        raise OAuthException(
            error=body.get("error", "Unknown Error"),
            error_description=body.get("error_description", ""),
        )
    return body


def _make_code_post(server, code, redirect_uri="urn:ietf:wg:oauth:2.0:oob"):
    data = {
        "client_id": CLIENT_ID,
        "client_secret": CLIENT_SECRET,
        "code": code,
        "grant_type": "authorization_code",
        "redirect_uri": redirect_uri,
    }
    info = _make_token_post(server, data)
    return info["access_token"], int(info["expires_in"]), info["refresh_token"]


def _make_refresh_post(server, refresh_token):
    data = {
        "client_id": CLIENT_ID,
        "client_secret": CLIENT_SECRET,
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    info = _make_token_post(server, data)
    return info["access_token"], int(info["expires_in"]), info["refresh_token"]


def _get_code(no_browser):
    """Make the requests to get OK access code"""

    host_name = REDIRECT_HOST
    try:
        port_number = _pick_free_port(port=REDIRECT_PORT)
    except OKException:
        # Could not bind to REDIRECT_HOST:0, try localhost instead
        host_name = "localhost"
        port_number = _pick_free_port(host_name, 0)

    redirect_uri = "http://{0}:{1}/".format(host_name, port_number)

    params = {
        "client_id": CLIENT_ID,
        "redirect_uri": redirect_uri,
        "response_type": "code",
        "scope": OAUTH_SCOPE,
    }
    url = "{}{}?{}".format(OK_SERVER_URL, AUTH_ENDPOINT, urlencode(params))

    server = OK_SERVER_URL
    code_response = None
    oauth_exception = None

    try:
        assert not no_browser
        assert webbrowser.open_new(url)
    except AssertionError:
        print("Couldn't open a web browser, performing manual authentication\n")
        print("Please navigate to https://go.cs61a.org/auth-code to generate a")
        print("login code. Then, paste that code below to finish logging in!\n")
        code = input("Paste login code here: ")
        try:
            code_response = _make_code_post(server, code)
            return code_response
        except OAuthException as e:
            raise e

    class CodeHandler(http.server.BaseHTTPRequestHandler):
        def send_redirect(self, location):
            self.send_response(302)
            self.send_header("Location", location)
            self.end_headers()

        def send_failure(self, oauth_exception):
            params = {
                "error": oauth_exception.error,
                "error_description": oauth_exception.error_description,
            }
            url = "{}{}?{}".format(server, ERROR_ENDPOINT, urlencode(params))
            self.send_redirect(url)

        def do_GET(self):
            """Respond to the GET request made by the OAuth"""
            nonlocal code_response, oauth_exception
            log.debug("Received GET request for %s", self.path)
            path = urlparse(self.path)
            qs = {k: v for k, v in parse_qsl(path.query)}
            code = qs.get("code")
            if code:
                try:
                    code_response = _make_code_post(server, code, redirect_uri)
                except OAuthException as e:
                    oauth_exception = e
            else:
                oauth_exception = OAuthException(
                    error=qs.get("error", "Unknown Error"),
                    error_description=qs.get("error_description", ""),
                )

            if oauth_exception:
                self.send_failure(oauth_exception)
            else:
                self.send_redirect(SUCCESS_ENDPOINT_URL)

        def log_message(self, format, *args):
            return

    server_address = (host_name, port_number)
    print("Authentication server running on {}:{}".format(host_name, port_number))

    try:
        httpd = http.server.HTTPServer(server_address, CodeHandler)
        httpd.handle_request()
    except OSError as e:
        log.warning("HTTP Server Err {}".format(server_address), exc_info=True)
        raise

    if oauth_exception:
        raise oauth_exception
    return code_response


class OAuthSession:
    """Represents OK OAuth state"""

    def __init__(self, access_token="", refresh_token="", expires_at=-1, session=None):
        """Create OK OAuth state with given tokens, and expiration"""
        self.session = self.refresh_token = self.access_token = None
        self.expires_at = -1
        self.assignment = None
        if session is not None:
            config = session.config()
            self.session = session
            if "ok_access_token" in config:
                self.access_token = config["ok_access_token"]
            if "ok_refresh_token" in config:
                self.refresh_token = config["ok_refresh_token"]
            if "ok_expires_at" in config:
                self.expires_at = int(config["ok_expires_at"])
            if "ok_last_download_assignment" in config:
                self.assignment = config["ok_last_download_assignment"]
        elif access_token and refresh_token and expires_at >= 0:
            self.access_token = str(access_token)
            self.refresh_token = str(refresh_token)
            self.expires_at = expires_at

    def _dump(self):
        """Dump state to a Bacon session"""
        if self.session is not None:
            config = self.session.config()
            if self.access_token:
                config["ok_access_token"] = self.access_token
            if self.refresh_token:
                config["ok_refresh_token"] = self.refresh_token
            if self.expires_at >= 0:
                config["ok_expires_at"] = str(self.expires_at)
            if self.assignment:
                config["ok_last_download_assignment"] = self.assignment

    def refresh(self):
        """Refreshes a token"""
        if not self.refresh_token:
            return False
        cur_time = int(time.time())
        if cur_time < self.expires_at - 3600:
            # expires in 1 hour
            return True
        self.access_token, expires_in, self.refresh_token = _make_refresh_post(
            OK_SERVER_URL, self.refresh_token
        )
        if not (self.access_token and expires_in):
            log.warning("Refresh authentication failed and returned an empty token.")
            return False
        cur_time = int(time.time())
        self.expires_at = cur_time + expires_in
        self._dump()
        return True

    def auth(self, force_reauth=False, no_browser=False):
        """
        Returns OAuth access token which can be passed to the server
        for identification. If force_reauth is specified then will
        force re-authenticate the user; else tries to reuse or
        refresh previous token
        """
        # Attempts to import SSL or raises an exception
        try:
            import ssl
        except:
            log.warning("Error importing SSL module", stack_info=True)
            sys.exit(1)
        else:
            log.info("SSL module is available")

        # Refresh the token if not forcing reauth
        if not force_reauth and self.refresh():
            return self.access_token

        # Perform OAuth
        print("Token is not available, performing OAuth")
        self.access_token, expires_in, self.refresh_token = _get_code(no_browser)
        return self.access_token


def set_token_path(path):
    global TOKEN_PATH
    TOKEN_PATH = abspath(path)


def refresh_token(no_browser=False):
    token = OAuthSession().auth(no_browser=no_browser)
    with open(TOKEN_PATH, "w+") as f:
        f.write(token)
    return token


def get_token():
    try:
        with open(TOKEN_PATH) as f:
            return f.read()
    except FileNotFoundError:
        return refresh_token()



from fpdf import FPDF

from examtool.api.assemble_export import AssembledExam, OptionQuestion, TextQuestion


def render_pdf_exam(assembled_exam: AssembledExam):
    pdf = FPDF()
    pdf.add_page()

    pdf.set_font("Courier", size=16)
    pdf.multi_cell(200, 20, txt=assembled_exam.exam, align="L")
    pdf.multi_cell(
        200,
        20,
        txt=assembled_exam.name.encode("latin-1", "replace").decode("latin-1"),
        align="L",
    )
    pdf.multi_cell(
        200,
        20,
        txt=assembled_exam.sid.encode("latin-1", "replace").decode("latin-1"),
        align="L",
    )

    pdf.set_font("Courier", size=9)

    def out(text):
        pdf.multi_cell(
            200,
            5,
            txt=text.encode("latin-1", "replace").decode("latin-1"),
            align="L",
        )

    for question in assembled_exam.questions:
        pdf.add_page()
        out("\nQUESTION")
        for line in question.prompt.text.split("\n"):
            out(line)

        out("\nANSWER")

        if isinstance(question, OptionQuestion):
            for option in question.options:
                if option in question.selected:
                    out("[X] " + option.text)
                else:
                    out("[ ] " + option.text)
        elif isinstance(question, TextQuestion):
            for line in question.response.split("\n"):
                out(line)
        else:
            assert False, f"Unknown question type {type(question)}"

        out("\nAUTOGRADER")
        out(question.autograde_output)

    return pdf



import click

from examtool.api.auth import refresh_token


@click.command()
@click.option(
    "--browser/--no-browser",
    default=True,
    help="Choose between browser-based and browserless authentication",
)
def login(browser):
    """
    Login to OKPy.
    """
    token = refresh_token(no_browser=not browser)
    print("Token = {}".format(token))
    print("Token automatically saved")


if __name__ == "__main__":
    login()



import argparse
import json
import socketserver
import ssl
import time
import traceback
import webbrowser
import os
from functools import wraps
from http import HTTPStatus, server
from http.server import HTTPServer
from urllib.error import URLError
from urllib.parse import unquote, urlparse, parse_qs
from urllib.request import Request, urlopen

STATIC_PATHS = {}
PATHS = {}

CONTENT_TYPE_LOOKUP = dict(
    html="text/html",
    css="text/css",
    js="application/javascript",
    svg="image/svg+xml",
)


def path_optional(decorator):
    def wrapped(func_or_path):
        if callable(func_or_path):
            return decorator("/" + func_or_path.__name__)(func_or_path)
        else:

            def actual_decorator(f):
                return decorator(func_or_path)(f)

            return actual_decorator

    return wrapped


def route(path):
    """Register a route handler."""

    if callable(path):
        return route("/" + path.__name__)(path)

    if not path.startswith("/"):
        path = "/" + path

    def wrap(f):
        if "." in path:
            STATIC_PATHS[path] = f
        else:
            PATHS[path] = f
        return f

    return wrap


class Handler(server.BaseHTTPRequestHandler):
    """HTTP handler."""

    def do_GET(self):
        try:
            self.send_response(HTTPStatus.OK)
            parsed_url = urlparse(unquote(self.path))
            path = parsed_url.path
            query_params = parse_qs(parsed_url.query)

            if path in STATIC_PATHS:
                out = bytes(STATIC_PATHS[path](**snakify(query_params)), "utf-8")
            else:
                path = GUI_FOLDER + path[1:]
                if "scripts" in path and not path.endswith(".js"):
                    path += ".js"
                if path == GUI_FOLDER:
                    path = GUI_FOLDER + "index.html"
                with open(path, "rb") as f:
                    out = f.read()

            self.send_header("Content-type", CONTENT_TYPE_LOOKUP[path.split(".")[-1]])
            self.end_headers()
            self.wfile.write(out)

        except Exception as e:
            print(e)

    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        raw_data = self.rfile.read(content_length).decode("utf-8")
        data = json.loads(raw_data)
        path = unquote(self.path)

        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "application/json")
        self.end_headers()

        try:
            result = PATHS[path](**snakify(data))
            self.wfile.write(bytes(json.dumps(result), "utf-8"))
        except Exception as e:
            print(e)
            raise

    def log_message(self, *args, **kwargs):
        pass


class Server:
    def __getattr__(self, item):
        def f(**kwargs):
            if IS_SERVER:
                return PATHS["/" + item](**kwargs)
            else:
                return multiplayer_post(item, kwargs)

        return f


Server = Server()


def multiplayer_post(path, data, server_url=None):
    """Post DATA to a multiplayer server PATH and return the response."""
    if not server_url:
        server_url = DEFAULT_SERVER
    data_bytes = bytes(json.dumps(data), encoding="utf-8")
    request = Request(server_url + "/" + path, data_bytes, method="POST")
    try:
        response = urlopen(request, context=ssl._create_unverified_context())
        text = response.read().decode("utf-8")
        if text.strip():
            return json.loads(text)
    except Exception as e:
        traceback.print_exc()
        print(e)
        return None


def multiplayer_route(path, server_path=None):
    """Convert a function that takes (data, send) into a route."""
    if not server_path:
        server_path = path

    def wrap(f):
        def send(data):
            return multiplayer_post(server_path, data)

        def routed_fn(data):
            response = f(data, send)
            return response

        route(path)(routed_fn)
        return f

    return wrap


@path_optional
def forward_to_server(path):
    def wrap(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            if IS_SERVER:
                return f(*args, **kwargs)
            else:
                return multiplayer_post(path, kwargs)

        return wrapped

    return wrap


def server_only(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        if IS_SERVER:
            return f(*args, **kwargs)
        else:
            raise Exception("Method not available locally!")

    return wrapped


def sendto(f):
    def wrapped(data):
        return f(**data)

    return wrapped


def start_server():
    global IS_SERVER
    IS_SERVER = True
    from flask import Flask, request, jsonify, send_from_directory, Response

    app = Flask(__name__, static_url_path="", static_folder="")
    for route, handler in PATHS.items():

        def wrapped_handler(handler=handler):
            return jsonify(handler(**snakify(request.get_json(force=True))))

        app.add_url_rule(route, handler.__name__, wrapped_handler, methods=["POST"])

    for route, handler in STATIC_PATHS.items():

        def wrapped_handler(route=route, handler=handler):
            query_params = parse_qs(request.query_string.decode())
            return Response(
                handler(**snakify(query_params)),
                mimetype=CONTENT_TYPE_LOOKUP[route.split(".")[-1]],
            )

        app.add_url_rule(route, handler.__name__, wrapped_handler, methods=["GET"])

    @app.route("/")
    def index():
        return send_from_directory("", "index.html")

    return app


def start_client(port, default_server, gui_folder, standalone):
    """Start web server."""
    global DEFAULT_SERVER, GUI_FOLDER, IS_SERVER
    DEFAULT_SERVER = default_server
    GUI_FOLDER = gui_folder
    IS_SERVER = False

    socketserver.TCPServer.allow_reuse_address = True
    httpd = HTTPServer(("localhost", port), Handler)
    if not standalone:
        webbrowser.open("http://localhost:" + str(port), new=0, autoraise=True)
    httpd.serve_forever()


def snakify(data):
    out = {}
    for key, val in data.items():
        snake_key = []
        for x in key:
            if x == x.upper():
                snake_key += "_"
            snake_key += x.lower()
        out["".join(snake_key)] = val
    return out


@route("/kill")
def kill():
    if not IS_SERVER:
        print("Exiting GUI")
        exit(0)


def start(port, default_server, gui_folder, db_init=None):
    global DEFAULT_SERVER
    DEFAULT_SERVER = default_server

    parser = argparse.ArgumentParser(description="Project GUI Server")
    parser.add_argument(
        "-s", help="Stand-alone: do not open browser", action="store_true"
    )
    parser.add_argument("-f", help="Force Flask app", action="store_true")
    args, unknown = parser.parse_known_args()

    import __main__

    if "gunicorn" not in os.environ.get("SERVER_SOFTWARE", "") and not args.f:
        request = Request(
            "http://127.0.0.1:{}/kill".format(port),
            bytes(json.dumps({}), encoding="utf-8"),
            method="POST",
        )
        try:
            urlopen(request)
            print("Killing existing gui process...")
            time.sleep(1)
        except URLError:
            pass

        start_client(port, default_server, gui_folder, args.s)
    else:
        if db_init:
            db_init()
        app = start_server()
        if args.f:
            app.run(port=port, threaded=False, processes=1)
        else:
            return app



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from tqdm import tqdm

from examtool.api.extract_questions import get_name
from examtool.api.grade import grade


@dataclass(frozen=True)
class AssembledExam:
    exam: str
    email: str
    name: str
    sid: str
    questions: List[Question]


@dataclass(frozen=True)
class Question:
    name: str
    prompt: Text
    autograde_output: str


@dataclass(frozen=True)
class OptionQuestion(Question):
    options: List[Text]
    selected: List[Text]


@dataclass(frozen=True)
class Text:
    text: str
    tex: str
    html: str

    def __init__(self, element):
        for k in ["text", "tex", "html"]:
            object.__setattr__(self, k, element[k])


@dataclass(frozen=True)
class TextQuestion(Question):
    response: str
    height: int


def assemble_exam(
    exam: str,
    email: Optional[str],
    response: Dict[str, Union[str, List[str]]],
    template_questions: List[Dict],
    student_questions: List[Dict],
    name_question: str,
    sid_question: str,
    dispatch,
    substitute_in_question_text: bool = False,
):
    questions = []

    exam = AssembledExam(
        exam=exam,
        email=email,
        name=response.get(name_question, "NO NAME"),
        sid=response.get(sid_question, "NO SID"),
        questions=questions,
    )

    student_question_lookup = {q["id"]: q for q in student_questions}

    for question in template_questions:
        question_name = get_name(question)

        if substitute_in_question_text:
            question_text = Text(student_question_lookup.get(question["id"], question))
        else:
            question_text = Text(question)

        autograde_output = (
            grade(
                email,
                student_question_lookup[question["id"]],
                response,
                dispatch,
            )
            if question["id"] in response and question["id"] in student_question_lookup
            else "STUDENT LEFT QUESTION BLANK"
            if question["id"] in student_question_lookup
            else "STUDENT DID NOT RECEIVE QUESTION"
        )

        if question.get("type") in ["multiple_choice", "select_all"]:
            selected_options = response.get(question["id"], [])
            if question.get("type") == "multiple_choice" and not isinstance(
                selected_options, list
            ):
                selected_options = [selected_options]

            available_options = [Text(option) for option in question["options"]]
            if question["id"] not in student_question_lookup:
                student_options = available_options
            else:
                student_options = [
                    option["text"]
                    for option in sorted(
                        student_question_lookup[question["id"]]["options"],
                        key=lambda option: option.get("index", ""),
                    )
                ]

            assert len(available_options) == len(student_options)

            assembled_question = OptionQuestion(
                name=question_name,
                prompt=question_text,
                options=available_options,
                selected=(
                    [
                        option
                        for option in available_options
                        if option.text in selected_options
                    ]
                ),
                autograde_output=autograde_output,
            )

        else:
            assembled_question = TextQuestion(
                name=question_name,
                prompt=question_text,
                autograde_output=autograde_output,
                response=response.get(question["id"], "").replace("\t", " " * 4),
                height=question.get("options") or 1,
            )

        questions.append(assembled_question)

    return exam


def export(
    template_questions,
    student_responses,
    exam,
    name_question,
    sid_question,
    *,
    dispatch=None,
    include_outline=True,
    substitute_in_question_text=False,
):
    assembles = {}

    if include_outline:
        assembles["OUTLINE"] = assemble_exam(
            exam,
            None,
            {},
            template_questions,
            template_questions,
            name_question,
            sid_question,
            dispatch,
        )

    for email, data in tqdm(student_responses.items(), desc="Assembling", unit="Exam"):
        assembles[email] = assemble_exam(
            exam,
            email,
            data.get("responses"),
            template_questions,
            data.get("student_questions"),
            name_question,
            sid_question,
            dispatch,
            substitute_in_question_text=substitute_in_question_text,
        )

    return assembles



import os
import pathlib
from io import BytesIO
from json import dump, load

import click
from pikepdf import Pdf

from examtool.api.convert import convert, load_imports
from examtool.api.database import get_exam
from examtool.api.gen_latex import render_latex
from examtool.api.scramble import scramble
from examtool.api.utils import sanitize_email
from examtool.cli.utils import (
    determine_semester,
    exam_name_option,
    hidden_output_folder_option,
    prettify,
)


@click.command()
@exam_name_option
@click.option(
    "--json",
    default=None,
    type=click.File("r"),
    help="The exam JSON you wish to compile. Leave blank to compile the deployed exam.",
)
@click.option(
    "--md",
    default=None,
    type=click.File("r"),
    help="The exam Markdown you wish to compile. Leave blank to compile the deployed exam.",
)
@click.option(
    "--seed",
    default=None,
    help="Scrambles the exam based off of the seed (E.g. a student's email).",
)
@click.option("--subtitle", prompt=False, default="Sample Exam.")
@click.option(
    "--with-solutions/--without-solutions",
    default=True,
    help="Generates the exam with (default) or without solutions.",
)
@click.option(
    "--exam-type",
    default="Final Exam",
    help="The type of exam you are given. For example 'Final Exam' (default).",
)
@click.option(
    "--semester",
    default=determine_semester(),
    help=f"The semester of the exam. '{determine_semester()}' (default).",
)
@click.option(
    "--json-out",
    default=None,
    type=click.File("w"),
    help="Exports the JSON to the file specified.",
)
@click.option(
    "--merged-md",
    default=None,
    type=click.File("w"),
    help="Merges any imports into a single file.",
)
@click.option(
    "--draft/--normal",
    default=False,
    help="Generates a draft copy of the exam, which is faster but less accurate.",
)
@click.option(
    "--num-threads",
    default=16,
    type=int,
    help="The number of threads to process the JSON file.",
)
@click.option(
    "--require-explicit-ids",
    default=False,
    is_flag=True,
    help="Raises an error if an ID is not specified by a question in its config.",
)
@hidden_output_folder_option
def compile(
    exam,
    json,
    md,
    seed,
    subtitle,
    with_solutions,
    exam_type,
    semester,
    json_out,
    merged_md,
    draft,
    num_threads,
    require_explicit_ids,
    out,
):
    """
    Compile one PDF or JSON (from Markdown), unencrypted.
    The exam may be deployed or local (in Markdown or JSON).
    If a seed is specified, it will scramble the exam.
    """
    if not out:
        out = ""

    pathlib.Path(out).mkdir(parents=True, exist_ok=True)

    if json:
        print("Loading exam...")
        exam_data = load(json)
    elif md:
        src = md.read()
        path = md.name
        if merged_md:
            buff = load_imports(src, path)
            merged_md.write("\n".join(buff.lines))
            return
        print("Compiling exam...")
        try:
            exam_data = convert(
                src,
                path=path,
                draft=draft,
                num_threads=num_threads,
                allow_random_ids=not require_explicit_ids,
            )
        except SyntaxError as e:
            print("SyntaxError:", e)
            print("Could not compile your exam!")
            return
    else:
        print("Fetching exam...")
        exam_data = get_exam(exam=exam)

    if seed:
        print("Scrambling exam...")
        exam_data = scramble(seed, exam_data, keep_data=with_solutions)

    def remove_solutions_from_groups(groups):
        for group in groups:
            # if isinstance(group, dict):
            group.pop("solution", None)
            if group.get("type") == "group":
                remove_solutions_from_groups(group.get("elements", []))

    if not seed and not with_solutions:
        print("Removing solutions...")
        groups = exam_data.get("groups", [])
        remove_solutions_from_groups(groups)

    if json_out:
        print("Dumping json...")
        dump(exam_data, json_out, indent=4, sort_keys=True)
        return

    print("Rendering exam...")
    settings = {
        "coursecode": prettify(exam.split("-")[0]),
        "description": subtitle,
        "examtype": exam_type,
        "semester": semester,
    }
    if seed:
        settings["emailaddress"] = sanitize_email(seed)
    with render_latex(
        exam_data,
        settings,
    ) as pdf:
        pdf = Pdf.open(BytesIO(pdf))
        pdf.save(os.path.join(out, exam + ".pdf"))
        pdf.close()


if __name__ == "__main__":
    compile()



import webbrowser

import click

from examtool.cli.utils import exam_name_option


@click.command()
@exam_name_option
@click.option(
    "--email", help="The email address of the student to impersonate.", prompt=True
)
def loginas(exam, email):
    """
    Login to examtool as a student.
    """
    webbrowser.open(f"https://exam.cs61a.org/{exam}#loginas={email}")


if __name__ == "__main__":
    loginas()



import csv
import os
from collections import namedtuple
from datetime import datetime
from os import getenv

import click
import pytz
import requests

TARGET = "https://docs.google.com/spreadsheets/d/1YAHnp7gwGwvfctDYIqFM3OFCQTlp_eY81lD2bq5how8/edit"
SHEET = "Exam #3"

AUTH_KEY = getenv("AUTH_KEY")
AUTH_SECRET = getenv("AUTH_SECRET")

EXAM_STARTS = {
    1588367400: "cs61a-friday-final",  # Fri 2:10PM
    1588410600: "cs61a-friday-final-alt",  # Sat 2:10AM
    1588392600: "cs61a-friday-final-alt-2",  # Fri 9:10PM
    1588399800: "cs61a-friday-final-alt-3",  # Fri 11:10PM
    1588417800: "cs61a-friday-final-alt-4",  # Sat 4:10AM
    1588443000: "cs61a-friday-final-alt-5",
}

Student = namedtuple("Student", ["exam", "deadline"])


def get_unix_timestamp(date_str, time_str):
    date = datetime.strptime(date_str + " " + time_str, "%a %m/%d %I:%M %p")
    date = date.replace(year=2020)
    date = pytz.timezone("America/Los_Angeles").localize(date)
    return int(date.timestamp())


@click.command()
@click.option("--out", default="rosters", prompt=True, type=click.Path())
def main(out):
    data = requests.post(
        "https://auth.apps.cs61a.org/google/read_spreadsheet",
        json={
            "url": TARGET,
            "sheet_name": SHEET,
            "client_name": AUTH_KEY,
            "secret": AUTH_SECRET,
        },
    ).json()

    headers = data[0]

    email_index = headers.index("Email")
    start_time_index = headers.index("Start Time")
    start_date_index = headers.index("Start Date")
    end_time_index = headers.index("End Time")
    end_date_index = headers.index("End Date")

    students = {}

    for row in data[1:]:
        if not row:
            continue
        start_time = get_unix_timestamp(row[start_date_index], row[start_time_index])
        end_time = get_unix_timestamp(row[end_date_index], row[end_time_index])

        if start_time not in EXAM_STARTS:
            raise Exception(
                "Unexpected start_time = {}, {} {}".format(
                    start_time, row[start_date_index], row[start_time_index]
                )
            )

        students[row[email_index]] = Student(
            exam=EXAM_STARTS[start_time], deadline=end_time
        )

    for exam in EXAM_STARTS.values():
        with open(
            os.path.join(out, exam.replace("-", "_").split("_", 1)[1] + "_roster.csv"),
            "w",
        ) as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerow(["Email", "Deadline"])
            for email, student in students.items():
                if student.exam == exam:
                    writer.writerow([email, student.deadline])
                else:
                    writer.writerow([email, 0])


if __name__ == "__main__":
    main()
